@@ -9,7 +9,7 @@
 # Notes:
 # - This is based on following online article:
 #       https://analyticsindiamag.com/step-by-step-guide-to-implement-multi-class-classification-with-bert-tensorflow
-# - Godawful code organization (e.g., double embedded functions).
+# - Godawful code organization (e.g., doubly embedded functions).
 # - Default text categories:
 #      Politics: 0
 #      Technology: 1
@@ -26,67 +26,81 @@
 #
 #  In our dataset we have text_a and label. The code will create objects for each of the above mentioned features for all the records in our dataset using the InputExample class provided in the BERT library.
 #
-# - To cache the tensorflow hub version of the BERT data, uses following steps,
+# - To cache the tensorflow hub version of the BERT data, use the following steps,
 #   based on https://medium.com/@xianbao.qian/how-to-run-tf-hub-locally-without-internet-connection-4506b850a915:
-#      local_hub_dir = /tmp/tf-hub
-#      mkdir local_hub_dir
-#      os.environ["TFHUB_CACHE_DIR"] = "/tmp/tf-hub"
-#      cache_subdir = hashlib.sha1(BERT_MODEL_HUB.encode("utf8")).hexdigest()
-#      tf_hub_url = BERT_MODEL_HUB.replace("https://tfhub.dev/", "https://storage.googleapis.com/tfhub-modules/")
-#      curl tf_hub_url /tmp/
-#      cd local_hub_dir
-#      tar xvfz /tmp/1.tar.gz
-#      move * cache_subdir
-#  For hub instance "https://tfhub.dev/google/bert_uncased_L-12_H-768_A-12/1",
-#  this downloads https://storage.googleapis.com/tfhub-modules/bert_uncased_L-12_H-768_A-12/1.tr.gz" and places the tar contents in /tmp/
+#       import hashlib
+#       from mezcla import glue_helpers as gh
+#       model = "uncased_L-12_H-768_A-12"
+#       # -or-: model = "uncased_L-8_H-512_A-8"
+#       BERT_MODEL_HUB = f"https://tfhub.dev/google/{model}/1"
+#       # take 2:
+#       #    model = "bert_uncased_L-8_H-512_A-8"
+#       #    BERT_MODEL_HUB = f"https://tfhub.dev/google/small_bert/{model}/2
+#       temp_dir = "/tmp"
+#       # -or-: temp_dir = f"{os.environ.get('HOME', '.')}/temp"
+#       local_hub_dir = f"{temp_dir}/tf-hub"
+#       gh.issue(f"mkdir {local_hub_dir}")
+#       os.environ["TFHUB_CACHE_DIR"] = local_hub_dir
+#       cache_subdir = hashlib.sha1(BERT_MODEL_HUB.encode("utf8")).hexdigest()
+#       gh.issue(f"mkdir {local_hub_dir}/{cache_subdir}")
+#       tf_hub_url = (BERT_MODEL_HUB.                                                        \
+#                         replace("https://tfhub.dev/google/",                               \
+#                                 "https://storage.googleapis.com/bert_models/2020_02_20/"). \
+#                         replace("/1", ".zip").replace("/2", ".zip").                       \
+#                         replace("/google", "").replace("/small_bert", "))
+#       temp_archive = f"{temp_dir}/{model}.zip"
+#       gh.issue(f"curl {tf_hub_url} > {temp_archive}")
+#       gh.issue(f"cd {local_hub_dir}/{cache_subdir}; unzip {temp_archive}")
+#       print([BERT_MODEL_HUB, local_hub_dir, cache_subdir, temp_archive])
+#   =>
+#   ['https://tfhub.dev/google/uncased_L-12_H-768_A-12/1', '/tmp/tf-hub', '627e8a1af062e84ad50e5e82c59b0f86c3a5f3df', '/tmp/uncased_L-12_H-768_A-12.zip']
+#   TODO: convert replace sequence to re.sub
+#
+#  For hub instance "https://tfhub.dev/google/uncased_L-12_H-768_A-12/1",
+#  this downloads https://storage.googleapis.com/tfhub-modules/2020_02_20/uncased_L-12_H-768_A-12.zip"
+#  and places the zipped contents in /tmp/tf-hub.
 #
 #--------------------------------------------------------------------------------
 # TODO:
 # - Determine the depencies between the versions of TensorFlow and BERT. This
-# doesn't work with TensorFlow 2.0 and BERT from 2018.
+#   doesn't work with TensorFlow 2.0 and BERT from 2018.
 #
 
 """Run text categorization over input file (or Predict the News Category Hackathon)"""
 
-## TODO: drop re import if not needed
+# Standard packages
 from datetime import datetime
+import hashlib
 import os
+
+# Installed packages
 import pandas as pd
-import re
-
-## OLD: from google.colab import drive
-
 import tensorflow as tf
 import tensorflow_hub as hub
 from sklearn.model_selection import train_test_split
-
-#Importing BERT modules
+# Import BERT modules
+# note: need to use bert-tenorflow version 1.0.1
 import bert
 from bert import run_classifier
-from bert import optimization
-from bert import tokenization
 
-from main import Main
-import debug
-import system
-import glue_helpers as gh
-
-# Get training data
-# TODO: replace with file specified on input
-## OLD: drive.mount("/GD")
+# Local packages
+from mezcla import debug
+from mezcla.main import Main
+from mezcla import system
+from mezcla import glue_helpers as gh
 
 # Show TensorFlow info
 # note: 1.15.0 and 0.7.0 for blog example
-print("tensorflow version : ", tf.__version__)
+print("tensorflow version : ", tf.__version__)                # pylint: disable=no-member
 print("tensorflow_hub version : ", hub.__version__)
 
 # Set the output directory for saving model file
-## OUTPUT_DIR = '/GD/My Drive/Colab Notebooks/BERT/bert_news_category'
 USER = system.getenv_text("USER", "user")
 BASE_DIR = system.getenv_text("BASE_DIR", ".")
-TRAINING_FILE = "Data_Train.xlsx"
-TESTING_FILE = "Data_Test.xlsx"
-TYPICAL_INPUT_DIR = system.form_path(USER, "data", "hackathon-news-prediction")
+TRAINING_FILE = "Data_Train.xls"
+TESTING_FILE = "Data_Test.xls"
+SUBMISSION_FILE = "submission_bert.xls"
+TYPICAL_INPUT_DIR = system.form_path("/home", USER, "data", "hackathon-news-prediction")
 DEFAULT_INPUT_DIR = "." if system.file_exists(TRAINING_FILE) else TYPICAL_INPUT_DIR
 INPUT_DIR = system.getenv_text("INPUT_DIR", DEFAULT_INPUT_DIR)
 TRAINING_PATH = system.form_path(INPUT_DIR, TRAINING_FILE)
@@ -94,12 +108,19 @@ TESTING_PATH = system.form_path(INPUT_DIR, TESTING_FILE)
 OUTPUT_DIR = system.getenv_text("OUTPUT_DIR",
                                 system.form_path(INPUT_DIR, "output"))
 SHOW_PLOTS = system.getenv_bool("SHOW_PLOTS", False)
-TFHUB_CACHE_DIR = system.getenv_text("TFHUB_CACHE_DIR")
+TFHUB_CACHE_DIR = system.getenv_text("TFHUB_CACHE_DIR", "/tmp/tf-hub",
+                                     "Directory for cache of models from Tensorflow Hub")
 if TFHUB_CACHE_DIR:
     debug.trace_fmt(3, "Using local TensorFlow Hub cache: {c}", c=TFHUB_CACHE_DIR)
+BERT_MODEL_HUB = system.getenv_text("BERT_MODEL_HUB", "https://tfhub.dev/google/bert_uncased_L-12_H-768_A-12/1",
+                                    "Model to use from Tensorflow Hub")
+USER_HOME = system.getenv_text("HOME", ".",
+                               "User home directory")
+## TODO
+## USER = system.getenv_text("USER", "joe",
+##                           "User ID")
 
-## #@markdown Whether or not to clear/delete the directory and create a new oneu
-## DO_DELETE = False #@param {type:"boolean"}
+# Whether or not to clear/delete the directory and create a new one
 DO_DELETE = system.getenv_bool("DO_DELETE", False)
 
 # Do sanity check on input files
@@ -107,35 +128,20 @@ debug.assertion(system.non_empty_file(TRAINING_PATH))
 debug.assertion(system.non_empty_file(TESTING_PATH))
 
 # Delete previous run if desired. The (re-)create the output directory if needed.
-
 debug.assertion(system.is_directory(INPUT_DIR))
 if DO_DELETE:
     try:
-        ## tf.gfile.DeleteRecursively(OUTPUT_DIR)
         gh.issue("/bin/rm --verbose --recursive {od}", od=OUTPUT_DIR)
     except:
-        ## pass
         system.print_stderr("Problem deleting dir: {od}", od=OUTPUT_DIR)
 if not system.is_directory(OUTPUT_DIR):
     system.create_directory(OUTPUT_DIR)
-
-## tf.gfile.MakeDirs(OUTPUT_DIR)
 print('***** Model output directory: {} *****'.format(OUTPUT_DIR))
     
-## TODO: Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
-## Note: Run following in Emacs to interactively replace TODO_ARGn with option label
-##    M-: (query-replace-regexp "todo\\([-_]\\)argn" "arg\\1name")
-## where M-: is the emacs keystroke short-cut for eval-expression.
-## TODO_ARG1 = False
-## TODO_ARG2 = "TODO-arg2"
-## TODO_FILENAME = "TODO-filename"
 
 class Script(Main):
-    """Input processing class"""
-    # TODO: -or-: """Adhoc script class (e.g., no I/O loop, just run calls)"""
-    BERT_MODEL_HUB = None
+    """Adhoc script class (e.g., no I/O loop, just run calls)"""
     # This is a path to an uncased (all lowercase) version of BERT
-    BERT_MODEL_HUB = "https://tfhub.dev/google/bert_uncased_L-12_H-768_A-12/1"
     MAX_SEQ_LENGTH = 128         # Sequences have at most 128 tokens.
     # These hyperparameters are copied from this colab notebook (https://colab.sandbox.google.com/github/tensorflow/tpu/blob/master/tools/colab/bert_finetuning_with_cloud_tpus.ipynb)
     BATCH_SIZE = 32
@@ -144,47 +150,67 @@ class Script(Main):
     # Warmup is a period of time where the learning rate is small and gradually increases--usually helps training.
     WARMUP_PROPORTION = 0.1
     # Model configs
-    SAVE_CHECKPOINTS_STEPS = 300
+    SAVE_CHECKPOINTS_STEPS = 1000
     SAVE_SUMMARY_STEPS = 100
-
+    #
     DATA_COLUMN = 'STORY'
     LABEL_COLUMN = 'SECTION'
     # The list containing all the classes (train['SECTION'].unique())
     LABEL_LIST = [0, 1, 2, 3]
+    #
+    tokenizer = None
+    estimator = None
 
-    # TODO: add class constructor
-    ## def __init__(self, *args, **kwargs):
-    ##     debug.trace_fmtd(5, "Script.__init__({a}): keywords={kw}; self={s}",
-    ##                      a=",".join(args), kw=kwargs, s=self)
-    ##     super(Script, self).__init__(*args, **kwargs)
-    
     def setup(self):
         """Check results of command line processing"""
         debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
-        ## TODO:
-        ## self.TODO_arg1 = self.get_parsed_option(TODO_ARG1, self.TODO_arg1)
-        ## self.TODO_arg2 = self.get_parsed_option(TODO_ARG2, self.TODO_arg2)
-        # TODO: self.TODO_filename = self.get_parsed_argument(TODO_FILENAME)
+        # Make CUDA settings (n.b., very idiosyncratic)
+        # Notes:
+        # 1. The actual library file are based on the GPU: see NVIDIA SDK.
+        # 2. Unfortunately, BERT requires Tensorflow 1.15 which uses old CUDA.
+        ## BAD: pkg_dir = gh.form_path(HOME, USER, ".conda", "pkgs")
+        pkg_dir = gh.form_path(USER_HOME, ".conda", "pkgs")
+        system.setenv("LD_LIBRARY_PATH", f"/usr/lib/x86_64-linux-gnu:/usr/lib/nvidia-450:/usr/local/cuda-11.1/lib64:{pkg_dir}/cudatoolkit-10.1.243-h6bb024c_0/lib:{pkg_dir}/cudnn-7.6.5-cuda10.1_0/lib:/usr/local/misc/lib/cuda-repo-ubuntu1604-10-0-local-10.0.130-410.48_1.0-1:/usr/local/cuda/extras/CUPTI/lib64")
+        # Show environment and this object instance
+        debug.trace(4, gh.run("printenv"))
         debug.trace_object(5, self, label="Script instance")
 
     def create_tokenizer_from_hub_module(self):
         """Get the vocab file and casing info from the Hub module."""
+        debug.trace(5, "create_tokenizer_from_hub_module()")
+        debug.reference_vars(self)
         with tf.Graph().as_default():
-            bert_module = hub.Module(self.BERT_MODEL_HUB)
+            bert_module = hub.Module(BERT_MODEL_HUB)
+            ## TODO
+            ## try:
+                # Take 1: Get from Tensorflow Hub server
+                ## bert_module = hub.Module(BERT_MODEL_HUB)
+            ## except:
+                ## raise NotImplementedError()
+                ## # Take 2: Get from cache (e.g., /home/tomohara/temp/tf-hub/3843e19587d721e7cfc9d9cfe564ca3da192ffc9)
+                ## # TODO: get the module conversion working
+                ## cache_subdir = hashlib.sha1(BERT_MODEL_HUB.encode("utf8")).hexdigest()
+                ## cache_path = os.path.join(TFHUB_CACHE_DIR, cache_subdir)
+                ## bert_module = hub.saved_model_module.create_module_spec_from_saved_model(cache_path)
             tokenization_info = bert_module(signature="tokenization_info", as_dict=True)
             with tf.Session() as sess:
                 vocab_file, do_lower_case = sess.run(
                     [tokenization_info["vocab_file"],
                      tokenization_info["do_lower_case"]])
-                
-        return bert.tokenization.FullTokenizer(
+                debug.trace_expr(5, bert_module, vocab_file, do_lower_case)
+        tokenizer = bert.tokenization.FullTokenizer(
             vocab_file=vocab_file, do_lower_case=do_lower_case)
+        debug.trace(4, f"create_tokenizer_from_hub_module() => {tokenizer}")
+        return tokenizer
+            
 
     def create_model(self, is_predicting, input_ids, input_mask, segment_ids,
                      labels, num_labels):
         """Load the BERT model for fine-tuning."""
+        debug.trace(5, f"create_model({tuple([is_predicting, input_ids, input_mask, segment_ids, labels, num_labels])})")
+        debug.reference_vars(self)
         bert_module = hub.Module(
-            self.BERT_MODEL_HUB,
+            BERT_MODEL_HUB,
             trainable=True)
         bert_inputs = dict(
             input_ids=input_ids,
@@ -224,22 +250,28 @@ class Script(Main):
   
             # If predicting, return predicted labels and the probabiltiies.
             if is_predicting:
-                return (predicted_labels, log_probs)
+                model = (predicted_labels, log_probs)
+                debug.trace(5, f"create_model() => {model}; first return")
+                return model
       
             # If in train/eval, compute loss between predicted and actual label
             # Note: returns loss, predicted labels, and the log of the label probs.
             per_example_loss = -tf.reduce_sum(one_hot_labels * log_probs, axis=-1)
             loss = tf.reduce_mean(per_example_loss)
-            return (loss, predicted_labels, log_probs)
+            model = (loss, predicted_labels, log_probs)
+            debug.trace(5, f"create_model() => {model}; second return")
+            return model
 
-    # model_fn_builder actually creates our model function
-    # using the passed parameters for num_labels, learning_rate, etc.
     def model_fn_builder(self, num_labels, learning_rate, num_train_steps,
                          num_warmup_steps):
         """Returns `model_fn` closure for TPUEstimator."""
+        # model_fn_builder actually creates our model function
+        # using the passed parameters for num_labels, learning_rate, etc.
+        debug.trace(5, f"model_fn_builder({tuple([num_labels, learning_rate, num_train_steps, num_warmup_steps])})")
 
         def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
             """The `model_fn` for TPUEstimator."""
+            debug.trace(5, f"model_fn({tuple([features, labels, mode, params])})")
             input_ids = features["input_ids"]
             input_mask = features["input_mask"]
             segment_ids = features["segment_ids"]
@@ -248,7 +280,7 @@ class Script(Main):
     
             # TRAIN and EVAL
             if not is_predicting:
-                (loss, predicted_labels, log_probs) = self.create_model(
+                (loss, predicted_labels, log_probs) = self.create_model(   # pylint: disable=unbalanced-tuple-unpacking
                     is_predicting, input_ids, input_mask, segment_ids, label_ids, num_labels)
     
                 train_op = bert.optimization.create_optimizer(
@@ -256,6 +288,8 @@ class Script(Main):
     
                 # Calculate evaluation metrics. 
                 def metric_fn(label_ids, predicted_labels):
+                    """Evaluate LABEL_IDS against PREDICTED_LABELS"""
+                    debug.trace(5, f"metric_fn({tuple([label_ids, predicted_labels])})")
                     accuracy = tf.metrics.accuracy(label_ids, predicted_labels)
                     true_pos = tf.metrics.true_positives(label_ids,
                                                          predicted_labels)
@@ -280,37 +314,37 @@ class Script(Main):
                     return tf.estimator.EstimatorSpec(mode=mode,
                                                       loss=loss,
                                                       train_op=train_op)
-                else:
-                    return tf.estimator.EstimatorSpec(mode=mode,
-                                                      loss=loss,
-                                                      eval_metric_ops=eval_metrics)
-            else:
-                (predicted_labels, log_probs) = self.create_model(
-                    is_predicting, input_ids, input_mask, segment_ids, label_ids, num_labels)
+                return tf.estimator.EstimatorSpec(mode=mode,
+                                                  loss=loss,
+                                                  eval_metric_ops=eval_metrics)
+            (predicted_labels, log_probs) = self.create_model(
+                is_predicting, input_ids, input_mask, segment_ids, label_ids, num_labels)
     
-                predictions = {
-                    'probabilities': log_probs,
-                    'labels': predicted_labels
-                }
-                return tf.estimator.EstimatorSpec(mode, predictions=predictions)
+            predictions = {
+                'probabilities': log_probs,
+                'labels': predicted_labels
+            }
+            return tf.estimator.EstimatorSpec(mode, predictions=predictions)
 
         # Return the actual model function in the closure
         return model_fn
 
     def getPrediction(self, in_sentences):
         """A method to get predictions"""
-        #A list to map the actual labels to the predictions
+        debug.trace(5, f"getPrediction({in_sentences})")
+
+        # A list to map the actual labels to the predictions
         labels = ["Politics", "Technology", "Entertainment", "Business"]
       
-        #Transforming the test data into BERT accepted form
+        # Transforming the test data into BERT accepted form
         input_examples = [run_classifier.InputExample(guid="", text_a=x, text_b=None, label=0) for x in in_sentences] 
         
-        #Creating input features for Test data
-        input_features = run_classifier.convert_examples_to_features(input_examples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, tokenizer)
+        # Creating input features for Test data
+        input_features = run_classifier.convert_examples_to_features(input_examples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)
       
-        #Predicting the classes 
+        # Predicting the classes 
         predict_input_fn = run_classifier.input_fn_builder(features=input_features, seq_length=self.MAX_SEQ_LENGTH, is_training=False, drop_remainder=False)
-        predictions = estimator.predict(predict_input_fn)
+        predictions = self.estimator.predict(predict_input_fn)
         return [(sentence, prediction['probabilities'], prediction['labels'], labels[prediction['labels']]) for sentence, prediction in zip(in_sentences, predictions)]
       
       
@@ -318,15 +352,12 @@ class Script(Main):
         """Main processing step"""
         debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)
 
-        # Loading The Data
-        ## OLD: We will now load the data from a Google Drive directory and will also split the training set in to training and validation sets.
-        ## train = pd.read_excel("/GD/My Drive/Colab Notebooks/News_category/Datasets/Data_Train.xlsx")
-        ## test = pd.read_excel("/GD/My Drive/Colab Notebooks/News_category/Datasets/Data_Test.xlsx")
+        # Load the data and also split the training set in to training and validation sets.
         train = pd.read_excel(system.form_path(INPUT_DIR, TRAINING_PATH))
         test = pd.read_excel(system.form_path(INPUT_DIR, TESTING_PATH))
-
         train, val =  train_test_split(train, test_size=0.2, random_state=100)
-        #Training set sample
+
+        # Show training set sample
         debug.trace_fmt(4, "Train sample:\n{s}", s=train.head(5))
         debug.trace_fmt(4, "Test sample:\n{s}", s=test.head(5))
         print("Training Set Shape :", train.shape)
@@ -348,126 +379,131 @@ class Script(Main):
             train['SECTION'].value_counts().plot(kind='bar')
 
         # Compute train and warmup steps from batch size
-        num_train_steps = int(len(train_features) / BATCH_SIZE * NUM_TRAIN_EPOCHS)
-        num_warmup_steps = int(num_train_steps * WARMUP_PROPORTION)
+        ## BAD: num_train_steps = int(len(train_features) / self.BATCH_SIZE * self.NUM_TRAIN_EPOCHS)
+        num_train_steps = int(len(train) / self.BATCH_SIZE * self.NUM_TRAIN_EPOCHS)
+        num_warmup_steps = int(num_train_steps * self.WARMUP_PROPORTION)
         
         # Specify output directory and number of checkpoint steps to save
         run_config = tf.estimator.RunConfig(
             model_dir=OUTPUT_DIR,
-            save_summary_steps=SAVE_SUMMARY_STEPS,
-            save_checkpoints_steps=SAVE_CHECKPOINTS_STEPS)
+            save_summary_steps=self.SAVE_SUMMARY_STEPS,
+            save_checkpoints_steps=self.SAVE_CHECKPOINTS_STEPS)
         
         # Specify output directory and number of checkpoint steps to save
         run_config = tf.estimator.RunConfig(
             model_dir=OUTPUT_DIR,
-            save_summary_steps=SAVE_SUMMARY_STEPS,
-            save_checkpoints_steps=SAVE_CHECKPOINTS_STEPS)
-        
-        
-        #Initializing the model and the estimator
+            save_summary_steps=self.SAVE_SUMMARY_STEPS,
+            save_checkpoints_steps=self.SAVE_CHECKPOINTS_STEPS)
+
+        # Initializing the model and the estimator
         model_fn = self.model_fn_builder(
             num_labels=len(self.LABEL_LIST),
-            learning_rate=LEARNING_RATE,
+            learning_rate=self.LEARNING_RATE,
             num_train_steps=num_train_steps,
             num_warmup_steps=num_warmup_steps)
-        
-        estimator = tf.estimator.Estimator(
+        #
+        self.estimator = tf.estimator.Estimator(
             model_fn=model_fn,
             config=run_config,
-            params={"batch_size": BATCH_SIZE})
+            params={"batch_size": self.BATCH_SIZE})
         
         # Read in training data
         # See 'Data preprocessing' notes in the header comments
         train_InputExamples = train.apply(lambda x:
                                           bert.run_classifier.InputExample(
                                               guid=None,
-                                              text_a=x[DATA_COLUMN], 
+                                              text_a=x[self.DATA_COLUMN], 
                                               text_b=None, 
-                                              label=x[LABEL_COLUMN]),
+                                              label=x[self.LABEL_COLUMN]),
                                           axis=1)
         val_InputExamples = val.apply(lambda x:
                                       bert.run_classifier.InputExample(
                                           guid=None, 
-                                          text_a=x[DATA_COLUMN], 
+                                          text_a=x[self.DATA_COLUMN], 
                                           text_b=None, 
-                                          label=x[LABEL_COLUMN]),
+                                          label=x[self.LABEL_COLUMN]),
                                       axis=1)
         debug.trace_object(5, train_InputExamples, "train_InputExamples")
-        debug.trace(4, "Row 0 - guid of training set: {r}", r=train_InputExamples.iloc[0].guid)
-        debug.trace(4, "__________\nRow 0 - text_a of training set {t}: ", r=train_InputExamples.iloc[0].text_a)
-        debug.trace(4, "__________\nRow 0 - text_b of training set: {t}", t=train_InputExamples.iloc[0].text_b)
-        debug.trace(4, "__________\nRow 0 - label of training set: {l}", l=train_InputExamples.iloc[0].label)
+        debug.trace_fmt(4, "Row 0 - guid of training set: {r}", r=train_InputExamples.iloc[0].guid)
+        debug.trace_fmt(4, "__________\nRow 0 - text_a of training set {t}: ", t=train_InputExamples.iloc[0].text_a)
+        debug.trace_fmt(4, "__________\nRow 0 - text_b of training set: {t}", t=train_InputExamples.iloc[0].text_b)
+        debug.trace_fmt(4, "__________\nRow 0 - label of training set: {l}", l=train_InputExamples.iloc[0].label)
 
         # Convert to BERT format
-        tokenizer = self.create_tokenizer_from_hub_module()
-        debug.trace_fmt(4, "row 0 tokenized: {t}", t=tokenizer.tokenize(train_InputExamples.iloc[0].text_a))
+        self.tokenizer = self.create_tokenizer_from_hub_module()
+        debug.trace_fmt(4, "row 0 tokenized: {t}", t=self.tokenizer.tokenize(train_InputExamples.iloc[0].text_a))
 
         # Convert train and validation features to InputFeatures that BERT understands.
         train_features = bert.run_classifier.convert_examples_to_features(
-            train_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, tokenizer)
+            train_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)
 
         val_features = bert.run_classifier.convert_examples_to_features(
-            val_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, tokenizer)
+            val_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)
         
         # Example on first observation in the training set
         debug.trace_fmt(4, "Sentence: {s}", s=train_InputExamples.iloc[0].text_a)
         debug.trace_fmt(4, "-"*30)
-        debug.trace_fmt(4, "Tokens: {t}", t=tokenizer.tokenize(train_InputExamples.iloc[0].text_a))
+        debug.trace_fmt(4, "Tokens: {t}", t=self.tokenizer.tokenize(train_InputExamples.iloc[0].text_a))
         debug.trace_fmt(4, "-"*30)
-        debug.trace_fmt(4, "Input IDs: {i}", l=train_features[0].input_ids)
+        debug.trace_fmt(4, "Input IDs: {i}", i=train_features[0].input_ids)
         debug.trace_fmt(4, "-"*30)
-        debug.trace_fmt(4, "Input Masks: {m}", m-train_features[0].input_mask)
+        debug.trace_fmt(4, "Input Masks: {m}", m=train_features[0].input_mask)
         debug.trace_fmt(4, "-"*30)
         debug.trace_fmt(4, "Segment IDs: {i}", i=train_features[0].segment_ids)
 
         # Initializing the model and the estimator
         model_fn = self.model_fn_builder(
             num_labels=len(self.LABEL_LIST),
-            learning_rate=LEARNING_RATE,
+            learning_rate=self.LEARNING_RATE,
             num_train_steps=num_train_steps,
             num_warmup_steps=num_warmup_steps)
 
-        estimator = tf.estimator.Estimator(
+        self.estimator = tf.estimator.Estimator(
             model_fn=model_fn,
             config=run_config,
-            params={"batch_size": BATCH_SIZE})
+            params={"batch_size": self.BATCH_SIZE})
 
-        # Create an input function for training. drop_remainder=True for using TPUs.
+        # Create an input function for training. Use drop_remainder=True when using TPUs.
         train_input_fn = bert.run_classifier.input_fn_builder(
             features=train_features,
             seq_length=self.MAX_SEQ_LENGTH,
             is_training=True,
             drop_remainder=False)
 
-        # Create an input function for validating. drop_remainder=True for using TPUs.
+        # Create an input function for validating. Use drop_remainder=True when using TPUs.
         val_input_fn = run_classifier.input_fn_builder(
             features=val_features,
             seq_length=self.MAX_SEQ_LENGTH,
             is_training=False,
             drop_remainder=False)
 
+        # Train the model
+        debug.trace(4, "Beginning Training!")
+        current_time = datetime.now()
+        self.estimator.train(input_fn=train_input_fn, max_steps=num_train_steps)
+        debug.trace_fmt(4, "Training took time {t}", t=(datetime.now() - current_time))
+        
         # Evaluating the model with Validation set
-        estimator.evaluate(input_fn=val_input_fn, steps=None)
+        self.estimator.evaluate(input_fn=val_input_fn, steps=None)
 
         # Evaluate over test set
+        # note: creates Excel submission file
         pred_sentences = list(test['STORY'])
-        predictions = getPrediction(pred_sentences)
-        debug.race_fmt(4, predictions[0])
-
-
+        predictions = self.getPrediction(pred_sentences)
+        debug.trace_expr(4, predictions[0])
+        #
         enc_labels = []
         act_labels = []
         for i in range(len(predictions)):
             enc_labels.append(predictions[i][2])
             act_labels.append(predictions[i][3])
-
-        ## OLD: pd.DataFrame(enc_labels, columns=['SECTION']).to_excel('/GD/My Drive/Colab Notebooks/BERT/submission_bert.xlsx', index=False)
-        data_file = system.form_path(INPUT_DIR, "submission_bert.xlsx")
+        #
+        data_file = system.form_path(INPUT_DIR, SUBMISSION_FILE)
         pd.DataFrame(enc_labels, columns=['SECTION']).to_excel(data_file,
                                                                index=False)
 
         # Classifying random sentences
-        tests = getPrediction(
+        tests = self.getPrediction(
             ['Mr.Modi is the Indian Prime Minister',
              'Gaming machines are powered by efficient micro processores and GPUs',
              'That HBO TV series is really good',
@@ -476,14 +512,6 @@ class Script(Main):
         debug.trace_object(4, tests, "tests")
         return
 
-        
-    ## TODO: def wrap_up(self):
-    ##           # ...
-
-    ## TODO: def clean_up(self):
-    ##           # ...
-    ##           super(Script, self).clean_up()
-
 #-------------------------------------------------------------------------------
     
 if __name__ == '__main__':
@@ -492,8 +520,5 @@ if __name__ == '__main__':
         description=__doc__,
         skip_input=False,
         manual_input=True,
-        ## boolean_options=[TODO_ARG1],
-        # TODO: positional_options=[TODO_FILENAME],
-        ## TODO: text_options=[(TODO_ARG2, "TODO-desc")]
     )
     app.run()
