@@ -17,21 +17,39 @@
 #   https://docs.python.org/3/tutorial/inputoutput.html
 #   https://www.python.org/dev/peps/pep-0498
 #
+# Usage examples:
+#
+#   - causal usage (quick-n-dirty debugging)
+#
+#     from mezcla import debug
+#     debug.trace_expr(4, dir())
+#
+#   - long-term usuage (e.g., symbolic constants for consistency across scripts)
+#
+#     from mezcla import debug
+#     TL = debug.TL
+#     debug.trace(TL.DEFAULT, "Shown by default")
+#     ...
+#     debug.trace(TL.USUAL, f"Look ma: TL.VERBOSE={int(TL.VERBOSE)}")
+#
+#   - use DEBUG_LEVEL env. var. to set level
+#
+#     python -c 'from mezcla import debug; debug.trace(debug.DEFAULT + 1, "Not visible")'
+#     DEBUG_LEVEL=3 python -c 'from mezcla import debug; debug.trace(3, "Visible")'
+#
 # TODO:
 # - * Add sanity checks for unused environment variables specified on command line (e.g., FUBAR=1 python script.py ...)!
 # - Rename as debug_utils so clear that non-standard package.
 # - Add exception handling throughout (e.g., more in trace_object).
+# - Apply format_value consistently.
 #
 
 """Debugging functions (e.g., tracing)"""
 
-## OLD: if sys.version_info.major == 2:
-## OLD:    from __future__ import print_function
-from __future__ import print_function
-
 # Standard packages
 import atexit
 from datetime import datetime
+import enum
 import inspect
 from itertools import zip_longest
 import logging
@@ -49,35 +67,61 @@ import mezcla.sys_version_info_hack      # pylint: disable=unused-import
 
 
 # Constants for pre-defined tracing levels
-ALWAYS = 0
-ERROR = 1
-WARNING = 2
-USUAL = 3
-DETAILED = 4
-VERBOSE = 5
-QUITE_DETAILED = 6
-QUITE_VERBOSE = 7
-MOST_DETAILED = 8
-MOST_VERBOSE = 9
+#
+class TraceLevel(enum.IntEnum):
+    """Constants for use in tracing"""
+    ALWAYS = 0
+    ERROR = 1
+    WARNING = 2                         # typically always shown
+    DEFAULT = WARNING
+    USUAL = 3                           # usual is sense of debugging purposes
+    DETAILED = 4
+    VERBOSE = 5
+    QUITE_DETAILED = 6
+    QUITE_VERBOSE = 7
+    MOST_DETAILED = 8
+    MOST_VERBOSE = 9
+#
+TL = TraceLevel
+ALWAYS = TL.ALWAYS
+ERROR = TL.ERROR
+WARNING = TL.WARNING
+DEFAULT = TL.DEFAULT
+USUAL = TL.USUAL
+DETAILED = TL.DETAILED
+VERBOSE = TL.VERBOSE
+QUITE_DETAILED = TL.QUITE_DETAILED
+QUITE_VERBOSE = TL.QUITE_VERBOSE
+MOST_DETAILED = TL.MOST_DETAILED
+MOST_VERBOSE = TL.MOST_VERBOSE
 
 # Other constants
 UTF8 = "UTF-8"
 STRING_TYPES = six.string_types
+INDENT = "    "
+
+# Globals
+max_trace_value_len = 512
 
 if __debug__:    
 
     # Initialize debug tracing level
     DEBUG_LEVEL_LABEL = "DEBUG_LEVEL"
-    trace_level = 1
+    trace_level = TL.DEFAULT            # typically same as TL.WARNING (2)
     output_timestamps = False           # prefix output with timestamp
+    last_trace_time = time.time()       # timestamp from last trace
     use_logging = False                 # traces via logging (and stderr)
     debug_file = None                   # file for log output
+    para_mode_tracing = False           # multiline tracing adds blank line (e.g., for para-mode grep)
     #
     try:
-        trace_level = int(os.environ.get(DEBUG_LEVEL_LABEL, trace_level))
+        trace_level_text = os.environ.get(DEBUG_LEVEL_LABEL, "")
+        if trace_level_text.strip():
+            trace_level = int(trace_level_text)
     except:
-        sys.stderr.write("Warning: Unable to set tracing level from {v}: {exc}\n".
-                         format(v=DEBUG_LEVEL_LABEL, exc=sys.exc_info()))
+        ## sys.stderr.write("Warning: Unable to set tracing level from {v}: {exc}\n".
+        ##                  format(v=DEBUG_LEVEL_LABEL, exc=sys.exc_info()))
+        pass
 
 
     def set_level(level):
@@ -89,9 +133,19 @@ if __debug__:
 
     def get_level():
         """Get current tracing level"""
-        ## global trace_level
-        return trace_level
-
+        # Note: ensures result is integer (not enum)
+        # EX: (get_level() >= 0)
+        # EX: type(get_level() == int)
+        ## OLD:
+        ## ## global trace_level
+        ## return trace_level
+        level = 0
+        try:
+            assertion(isinstance(level, int))
+            level = int(trace_level)
+        except:
+            _print_exception_info("get_level")
+        return level
 
     def get_output_timestamps():
         """Return whether outputting timestamps"""
@@ -149,7 +203,11 @@ if __debug__:
             if output_timestamps:
                 # Get time-proper from timestamp (TODO: find standard way to do this)
                 timestamp_time = re.sub(r"^\d+-\d+-\d+\s*", "", timestamp())
-
+                if detailed_debugging():
+                    global last_trace_time
+                    diff = round(1000.0 * (time.time() - last_trace_time), 3)
+                    timestamp_time += f" diff={diff}ms"
+                    last_trace_time = time.time()
                 print("[" + timestamp_time + "]", end=": ", file=sys.stderr)
                 if debug_file:
                     print("[" + timestamp_time + "]", end=": ", file=debug_file)
@@ -200,25 +258,28 @@ if __debug__:
 
 
     STANDARD_TYPES = (int, float, dict, list)
+    SIMPLE_TYPES = (bool, int, float, type(None), str)
     #
-    MAX_OBJECT_VALUE_LEN = 128
-    #
-    def trace_object(level, obj, label=None, show_all=False, indentation=None, pretty_print=None, max_depth=0):
+    def trace_object(level, obj, label=None, show_all=None, show_private=None, show_methods_etc=None, indentation=None, pretty_print=None, max_value_len=max_trace_value_len, max_depth=0):
         """Trace out OBJ's members to stderr if at trace LEVEL or higher.
         Note: Optionally uses output LABEL, with INDENTATION, SHOWing_ALL members, and PRETTY_PRINTing.
+        TODO: Use SHOW_PRIVATE to display private members and SHOW_METHODS_ETC for methods.
         If max_depth > 0, this uses recursion to show values for instance members."""
         # HACK: Members for STANDARD_TYPES omitted unless show_all.
-        # Note: This is intended for arbitrary objects, use trace_values for objects known to be lists or hashes.
-        # See https://stackoverflow.com/questions/383944/what-is-a-python-equivalent-of-phps-var-dump.
+        # Notes:
+        # - This is intended for arbitrary objects, use trace_values for objects known to be lists or hashes.
+        # - Support for show_private and show_methods_etc is not yet implemented (added for sake of tpo_common.py).
+        # - See https://stackoverflow.com/questions/383944/what-is-a-python-equivalent-of-phps-var-dump.
         # TODO: support recursive trace; specialize show_all into show_private and show_methods
         ## OLD: print("{stmt} < {current}: {r}".format(stmt=level, current=trace_level,
+        # TODO: handle tuples
         ##                                       r=(trace_level < level)))
-        trace_fmt(10, "trace_object({dl}, {obj}, label={lbl}, show_all={sa}, indent={ind}, pretty={pp}, max_d={md})",
+        trace_fmt(MOST_VERBOSE, "trace_object({dl}, {obj}, label={lbl}, show_all={sa}, indent={ind}, pretty={pp}, max_d={md})",
                   dl=level, obj=object, lbl=label, sa=show_all, ind=indentation, pp=pretty_print, md=max_depth)
-        if (pretty_print is None):
-            pretty_print = (trace_level > level)
         if (trace_level < level):
             return
+        if (pretty_print is None):
+            pretty_print = (trace_level > level)
         type_id_label = str(type(obj)) + " " + hex(id(obj))
         if label is None:
             ## BAD: label = str(type(obj)) + " " + hex(hash(obj))
@@ -228,128 +289,149 @@ if __debug__:
             label += " [" + type_id_label + "]"
         else:
             pass
+        outer_indentation = ""
         if indentation is None:
-            indentation = "   "
-        trace(0, label + ": {")
+            indentation = INDENT
+        else:
+            # TODO: rework via indent_level arg
+            outer_indentation = indentation[:len(indentation)-len(INDENT)]
+        if show_all is None:
+            show_all = (show_private or show_methods_etc)
+        if para_mode_tracing:
+            trace(ALWAYS, "")
+        trace(ALWAYS, outer_indentation + label + ": {")
         ## OLD: for (member, value) in inspect.getmembers(obj):
         member_info = []
         try:
             member_info = inspect.getmembers(obj)
         except:
-            trace_fmtd(7, "Warning: Problem getting member list in trace_object: {exc}",
+            trace_fmtd(QUITE_VERBOSE, "Warning: Problem getting member list in trace_object: {exc}",
                        exc=sys.exc_info())
         ## HACK: show standard type value as special member
         if isinstance(obj, STANDARD_TYPES):
             member_info = [("(value)", obj)] + [(("__(" + m + ")__"), v) for (m, v) in member_info]
-            trace_fmtd(7, "{ind}Special casing standard type as member {m}",
+            trace_fmtd(QUITE_VERBOSE, "{ind}Special casing standard type as member {m}",
                        ind=indentation, m=member_info[0][0])
         for (member, value) in member_info:
             # If high trace level, output the value as is
             # TODO: value = clip_text(value)
-            trace_fmtd(8, "{i}{m}={v}; type={t}", i=indentation, m=member, v=value, t=type(value))
-            if (trace_level >= 9):
-                sys.stderr.write(indentation + member + ":")
+            trace_fmtd(MOST_DETAILED, "{i}{m}={v}; type={t}", i=indentation, m=member, v=value, t=type(value))
+            value_spec = format_value(value, max_len=max_value_len)
+            if (trace_level >= MOST_VERBOSE):
+                sys.stderr.write(indentation + member + ": ")
                 if pretty_print:
-                    pprint(value, stream=sys.stderr)
+                    pprint(value_spec, stream=sys.stderr)
                 else:
-                    sys.stderr.write(value, file=sys.stderr)
+                    sys.stderr.write(value_spec)
                 sys.stderr.write("\n")
                 if use_logging:
-                    logging.debug(_to_utf8((indentation + member + ":" + str(value))))
+                    logging.debug(_to_utf8((indentation + member + ": " + value_spec)))
                 continue
-            # Optionally, process recursively (TODO: make "   " an env. option)
-            if (max_depth > 0):
-                trace_object(level, value, label=member, show_all=show_all,
-                             indentation=(indentation + "   "), pretty_print=None,
+            # Include unless special member (or if no filtering)
+            ## DEBUG: trace_expr(QUITE_VERBOSE, member.startswith("__"), re.search(r"^<.*(method|module|function).*>$", value_spec))
+            include_member = (show_all or (not (member.startswith("__") or 
+                                                re.search(r"^<.*(method|module|function).*>$", value_spec))))
+            # Optionally, process recursively (TODO: make INDENT an env. option)
+            if ((max_depth > 0) and include_member and (not isinstance(value, SIMPLE_TYPES))):
+                # TODO: add helper for formatting type & address (for use here and above)
+                member_type_id_label = (member + " [" + str(type(value)) + " " + hex(id(value)) + "]")
+                trace_object(level, value, label=member_type_id_label, show_all=show_all,
+                             indentation=(indentation + INDENT), pretty_print=None,
                              max_depth=(max_depth - 1))
                 continue
             # Otherwise, derive value spec. (trapping for various exceptions)
             ## TODO: pprint.pprint(member, stream=sys.stderr, indent=4, width=512)
             try:
                 try:
-                    value_spec = "%s" % ((value),)
-                    if (len(value_spec) > MAX_OBJECT_VALUE_LEN):
-                        value_spec = value_spec[:MAX_OBJECT_VALUE_LEN] + "..."
+                    value_spec = format_value("%s" % ((value),), max_len=max_value_len)
                 except(TypeError, ValueError):
-                    trace_fmtd(7, "Warning: Problem in tracing member {m}: {exc}",
+                    trace_fmtd(QUITE_VERBOSE, "Warning: Problem in tracing member {m}: {exc}",
                                m=member, exc=sys.exc_info())
                     value_spec = "__n/a__"
             except(AttributeError):
                 # Note: This can occur when profile_function set
-                trace_fmtd(7, "Error: unexpected problem in trace_object: {exc}",
+                trace_fmtd(QUITE_VERBOSE, "Error: unexpected problem in trace_object: {exc}",
                            exc=sys.exc_info())
                 value_spec = "__n/a__"
-            # Output unless special member (or if no filtering)
-            if (show_all or (not (member.startswith("__") or 
-                                  re.search(r"^<.*(method|module|function).*>$", value_spec)))):
-                ## trace(0, indentation + member + ": " + value_spec)
+            if include_member:
                 sys.stderr.write(indentation + member + ": ")
                 if pretty_print:
                     # TODO: remove quotes from numbers and booleans
                     pprint(value_spec, stream=sys.stderr, indent=len(indentation))
                 else:
-                    ## sys.stderr.write(value_spec)
                     sys.stderr.write(_to_utf8(value_spec))
                     sys.stderr.write("\n")
                 if use_logging:
                     logging.debug(_to_utf8((indentation + member + ":" + value_spec)))
-        trace(0, indentation + "}")
+        trace(ALWAYS, indentation + "}")
+        if para_mode_tracing:
+            trace(ALWAYS, "")
         return
 
 
     def trace_values(level, collection, label=None, indentation=None, use_repr=None):
         """Trace out elements of array or hash COLLECTION if at trace LEVEL or higher"""
-        trace_fmt(10, "trace_values(dl, {coll}, label={lbl}, indent={ind})",
+        trace_fmt(MOST_VERBOSE, "trace_values(dl, {coll}, label={lbl}, indent={ind})",
                   dl=level, lbl=label, coll=collection, ind=indentation)
         if (trace_level < level):
             return
-        if hasattr(collection, '__iter__'):
-             trace(level + 1, "Warning: [trace_values] consuming iterator")
-             collection = list(collection)             
+        if para_mode_tracing:
+            trace(ALWAYS, "")
         if not isinstance(collection, (list, dict)):
-            trace(level + 1, "Warning: [trace_values] coercing input into list")
-            collection = list(collection)
+            if hasattr(collection, '__iter__'):
+                trace(level + 1, "Warning: [trace_values] consuming iterator")
+                collection = list(collection)
+            else:
+                trace(level + 1, "Warning: [trace_values] coercing input into list")
+                collection = list(collection)
         if indentation is None:
-            indentation = "   "
+            indentation = INDENT
         if label is None:
             ## BAD: label = str(type(collection)) + " " + hex(hash(collection))
             label = str(type(collection)) + " " + hex(id(collection))
-            indentation = "   "
+            indentation = INDENT
         if use_repr is None:
             use_repr = False
-        trace(0, label + ": {")
+        trace(ALWAYS, label + ": {")
         keys_iter = list(collection.keys()) if isinstance(collection, dict) else range(len(collection))
         for k in keys_iter:
             try:
-                value = _to_utf8(collection[k])
+                value = format_value(_to_utf8(collection[k]))
                 if use_repr:
                     value = repr(value)
-                trace_fmtd(0, "{ind}{k}: {v}", ind=indentation, k=k,
+                trace_fmtd(ALWAYS, "{ind}{k}: {v}", ind=indentation, k=k,
                            v=value)
             except:
-                trace_fmtd(7, "Warning: Problem tracing item {k}",
+                trace_fmtd(QUITE_VERBOSE, "Warning: Problem tracing item {k}",
                            k=_to_utf8(k), exc=sys.exc_info())
-        trace(0, indentation + "}")
+        trace(ALWAYS, indentation + "}")
+        if para_mode_tracing:
+            trace(ALWAYS, "")
         return
 
 
     def trace_expr(level, *values, **kwargs):
         """Trace each of the arguments (if at trace LEVEL or higher), using introspection
-        to derive label for each expression;
+        to derive label for each expression. By default, the following format is used:
+           expr1=value1; ... exprN=valueN
         Notes:
         - For simplicity, the values are assumed to separated by ', ' (or expression _SEP)--barebones parsing applied.
-        - Use DELIM to specify delimiter; otherwise \n used;
+        - Use DELIM to specify delimiter; otherwise '; ' used;
           if so, NO_EOL applies to intermediate values (EOL always used at end).
         - Use USE_REPR=False to use tracing via str instead of repr.
         - Use _KW_ARG for KW_ARG in case of conflict, as in following:
-          trace_expr(4, term, _term="; ")
+          trace_expr(DETAILED, term, _term="; ")
+        - Use MAX_LEN to specify maximum value length.
+        - Use PREFIX to specify initial trace output
         - See misc_utils.trace_named_objects for similar function taking string input, which is more general but harder to use and maintain"""
-        trace_fmt(10, "trace_expr({l}, a={args}, kw={kw})",
+        trace_fmt(MOST_VERBOSE, "trace_expr({l}, a={args}, kw={kw})",
                   l=level, args=values, kw=kwargs)
         sep = kwargs.get('sep') or kwargs.get('_sep')
         delim = kwargs.get('delim') or kwargs.get('_delim')
         no_eol = kwargs.get('no_eol') or kwargs.get('_no_eol')
         use_repr = kwargs.get('use_repr') or kwargs.get('_use_repr')
+        max_len = kwargs.get('max_len') or kwargs.get('_max_len')
+        prefix = kwargs.get('prefix') or kwargs.get('_prefix')
         if sep is None:
             sep = ", "
         if no_eol is None:
@@ -359,6 +441,8 @@ if __debug__:
             no_eol = True
         if use_repr is None:
             use_repr = True
+        if prefix is None:
+            prefix = ""
         # Get symbolic expressions for the values
         # TODO: handle cases split across lines
         try:
@@ -376,17 +460,24 @@ if __debug__:
             # Skip first argument (level)
             expressions = statement.split(sep)[1:]
         except:
-            trace_fmtd(0, "Exception isolating expression in trace_vals: {exc}",
+            trace_fmtd(ALWAYS, "Exception isolating expression in trace_vals: {exc}",
                        exc=sys.exc_info())
             expressions = []
+        trace(level, prefix, no_eol=no_eol)
         for expression, value in zip_longest(expressions, values):
-            # Exclude kwarg params
-            match = re.search(r"^(\w+)=", expression)
-            if (match and match.group(1) in kwargs):
-                continue
-            assertion(not (value and expression is None), "Warning: Likely problem resolving expression text (try reworking trace_expr call)")
-            value_spec = repr(value) if use_repr else value
-            trace(level, f"{expression}={value_spec}{delim}", no_eol=no_eol)
+            try:
+                # Exclude kwarg params
+                match = re.search(r"^(\w+)=", str(expression))
+                if (match and match.group(1) in kwargs):
+                    continue
+                assertion((not ((value is not None) and (expression is None))),
+                          "Warning: Likely problem resolving expression text (try reworking trace_expr call)")
+                value_spec = format_value(repr(value) if use_repr else value,
+                                          max_len=max_len)
+                trace(level, f"{expression}={value_spec}{delim}", no_eol=no_eol)
+            except:
+                trace_fmtd(ALWAYS, "Exception tracing values in trace_vals: {exc}",
+                       exc=sys.exc_info())            
         if no_eol:
             trace(level, "", no_eol=False)
         return
@@ -405,10 +496,12 @@ if __debug__:
         try:
             frame = inspect.currentframe().f_back
         except (AttributeError, KeyError, ValueError):
-            trace_fmt(5, "Exception during trace_current_context: {exc}",
+            trace_fmt(VERBOSE, "Exception during trace_current_context: {exc}",
                       exc=sys.exc_info())
+        if para_mode_tracing:
+            trace(level, "")
         trace_fmt(level, "{label} context: {{", label=label)
-        prefix = "    "
+        prefix = INDENT
         if (get_level() - level) > 1:
             trace_object((level + 2), frame, "frame", indentation=prefix,
                          show_all=show_methods_etc)
@@ -420,6 +513,8 @@ if __debug__:
                 trace_object(level, frame.f_locals, "locals", indentation=prefix,
                              show_all=show_methods_etc)
         trace(level, "}")
+        if para_mode_tracing:
+            trace(level, "")
         return
 
 
@@ -433,7 +528,7 @@ if __debug__:
     def trace_exception(level, task):
         """Trace exception information regarding TASK (e.g., function) at LEVEL"""
         # Note: Conditional output version of system's print_exception_info.
-        # ex: trace_exception(4, "tally_counts")
+        # ex: trace_exception(DETAILED, "tally_counts")
         trace_fmt(level, "Exception during {t}: {exc}".
                   format(t=task, exc=sys.exc_info()))
         # TODO: include full stack trace (e.g., at LEVEL + 2)
@@ -451,14 +546,15 @@ if __debug__:
 
 
     def assertion(expression, message=None):
-        """Issue warning if EXPRESSION doesn't hold, along with optional MESSAGE"""
+        """Issue warning if EXPRESSION doesn't hold, along with optional MESSAGE
+        Note: This is a "soft assertion" that doesn't raise an exception (n.b., provided the test doesn't do so)"""
         # EX: assertion((2 + 2) != 5)
         # TODO: have streamlined version using sys.write that can be used for trace and trace_fmtd sanity checks about {}'s
-        # TODO: trace out local and globals to aid in diagnosing assertion failures
+        # TODO: trace out local and globals to aid in diagnosing assertion failures; ex: add automatic tarcing of variables used in the assertion expression)
         if (not expression):
             try:
                 # Get source information for failed assertion
-                trace_fmtd(9, "Call stack: {st}", st=inspect.stack())
+                trace_fmtd(MOST_VERBOSE, "Call stack: {st}", st=inspect.stack())
                 caller = inspect.stack()[1]
                 (_frame, filename, line_number, _function, _context, _index) = caller
                 # Read statement in file and extract assertion expression
@@ -470,39 +566,52 @@ if __debug__:
                 qualification_spec = (": " + message) if message else ""
                 # Output information
                 # TODO: omit subsequent warnings
-                trace_fmtd(0, "Assertion failed: {expr} (at {file}:{line}){qual}",
+                trace_fmtd(ALWAYS, "Assertion failed: {expr} (at {file}:{line}){qual}",
                            expr=expression, file=filename, line=line_number, qual=qualification_spec)
             except:
-                trace_fmtd(0, "Exception formatting assertion: {exc}",
+                trace_fmtd(ALWAYS, "Exception formatting assertion: {exc}",
                            exc=sys.exc_info())
-                trace_object(0, inspect.currentframe(), "caller frame", pretty_print=True)
+                trace_object(ALWAYS, inspect.currentframe(), "caller frame", pretty_print=True)
         return
 
+    def val(level, value):
+        """Returns VALUE if at trace LEVEL or higher otherwise None
+        Note: inspired by Lisp's convenient IF form without an explicit else: (if test value-if-true)"""
+        # EX: (101 if ((get_level() == 1) and val(1, 101)) else None) => 101
+        # EX: ((not __debug__) and val(trace_level, 101))) => None
+        # TODO: rename as cond_value???
+        return (value if (trace_level >= level) else None)
+
     def code(level, no_arg_function):
         """Execute NO_ARG_FUNCTION if at trace LEVEL or higher
-        Note: Thanks to the wonders of Python syntax, a two-step process is required:
+        Notes:
+        - Given the quirks of Python syntax, a two-step process is required:
            debug.code(4, { line1; line2; ...; lineN })
-        =>
-           if __debug__:
-               def my_stupid_block_workaround(): 
+               =>
+           def my_stupid_block_workaround(): 
+               if __debug__:
                    line1; line2; ...; lineN
-           debug.code(4, my_stupid_block_workaround)"""
-        trace_object(5, f"code({level}, {no_arg_function})")
-        if level:
-            trace_object(6, f"Executing {no_arg_function}")
+           debug.code(4, my_stupid_block_workaround)
+        - Lambda functions can be used for simple expression-based functions"""
+        trace_object(VERBOSE, f"code({level}, {no_arg_function})")
+        if (trace_level >= level):
+            trace_object(QUITE_DETAILED, f"Executing {no_arg_function}")
             no_arg_function()
         return
 
 else:
 
+    trace_level = 0
+    
     def non_debug_stub(*_args, **_kwargs):
-        """Non-debug stub"""
+        """Non-debug stub (i.e., no-op function)"""
+        # Note: no return value assumed by debug.expr
         return
 
 
     def get_level():
         """Returns tracing level (i.e., 0)"""
-        return 0
+        return trace_level
 
 
     def get_output_timestamps():
@@ -531,6 +640,22 @@ else:
     assertion = non_debug_stub
 
     code = non_debug_stub
+
+    ## TODO?:
+    ## val = non_debug_stub
+    ##
+    def val(_expression):
+        """Non-debug stub for value()--a no-op function"""
+        # Note: implemented separately from non_debug_stub to ensure no return value
+        return
+    ##
+    assert val(1) is None, "non-debug val() should not return a non-Null value"
+
+# Aliases for terse functions
+    
+cond_code = code
+
+cond_val = val
     
 # note: adding alias for trace_fmtd to account for common typo
 # TODO: alias trace to trace_fmt as well (add something like trace_out if format not desired)
@@ -569,20 +694,45 @@ def verbose_debugging():
 
 
 def _getenv_bool(name, default_value):
-    """Version of debug.getenv_bool w/o tracing"""
+    """Version of system.getenv_bool w/o tracing"""
     result = default_value
     if (str(os.environ.get(name) or default_value).upper() in ["1", "TRUE"]):
         result = True
     return result
 
 
+def _getenv_int(name, default_value):
+    """Version of system.getenv_int w/o tracing"""
+    result = default_value
+    try:
+        env_value = os.environ.get(name)
+        if env_value is not None:
+            result = int(env_value)
+    except:
+        _print_exception_info("_getenv_int")
+    return result
+
+
+def format_value(value, max_len=None):
+    """Format VALUE for output with trace_values, etc.: truncates if too long and encodes newlines"""
+    # EX: format_value("    \n\n\n\n", 6) => "    \\n\\n..."
+    trace(MOST_VERBOSE, f"format_value({value}, max_len={max_len})")
+    if max_len is None:
+        max_len = max_trace_value_len
+    result = value if isinstance(value, str) else str(value)
+    if len(result) > max_len:
+        result = result[:max_len] + "..."
+    result = re.sub("\n", r"\\n", result)
+    return result
+
+
 def xor(value1, value2):
     """Whether VALUE1 and VALUE2 differ when interpretted as booleans"""
     # Note: Used to clarify assertions; same as bool(value1) != bool(value2).
     # See https://stackoverflow.com/questions/432842/how-do-you-get-the-logical-xor-of-two-variables-in-python
     # EX: not(xor(0, 0.0))
     result = (bool(value1) ^ bool(value2))
-    trace_fmt(7, "xor({v1}, {v2}) => {r}", v1=value1, v2=value2, r=result)
+    trace_fmt(QUITE_VERBOSE, "xor({v1}, {v2}) => {r}", v1=value1, v2=value2, r=result)
     return result
 
 
@@ -592,27 +742,27 @@ def xor3(value1, value2, value3):
     ##           and not (bool(value1) and bool(value2) and bool(value3)))
     num_true = sum([int(bool(v)) for v in [value1, value2, value3]])
     result = (num_true == 1)
-    trace_fmt(7, "xor3({v1}, {v2}, {v3}) => {r}",
+    trace_fmt(QUITE_VERBOSE, "xor3({v1}, {v2}, {v3}) => {r}",
               v1=value1, v2=value2, v3=value3, r=result)
     return result
 
 
 def init_logging():
     """Enable logging with INFO level by default or with DEBUG if detailed debugging"""
-    trace(4, "init_logging()")
-    trace_object(6, logging.root, "logging.root")
+    trace(DETAILED, "init_logging()")
+    trace_object(QUITE_DETAILED, logging.root, "logging.root")
 
     # Set the level for the current module
     # TODO: use mapping from symbolic LEVEL user option (e.g., via getenv)
     level = logging.DEBUG if detailed_debugging() else logging.INFO
-    trace_fmt(5, "Setting logger level to {ll}", ll=level)
+    trace_fmt(VERBOSE, "Setting logger level to {ll}", ll=level)
     logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=level)
     logging.debug("init_logging()")
 
     # Optionally make sure logging level applied globally
     if _getenv_bool("GLOBAL_LOGGING", False):
         old_level = logging.root.level
-        trace_fmt(5, "Setting root logger level from {ol} to {nl}", ol=old_level, nl=level)
+        trace_fmt(VERBOSE, "Setting root logger level from {ol} to {nl}", ol=old_level, nl=level)
         logging.root.setLevel(level)
     return
 
@@ -637,8 +787,8 @@ def profile_function(frame, event, arg):
     #    of the object in the local namespace corresponding to "self".
     # TODO:
     # - Make tracing level an option (e.g., environment).
-    trace_fmt(6, "profile_function(_, {e}, {a})", e=event, a=arg)
-    trace_object(7, frame, "frame")
+    trace_fmt(QUITE_DETAILED, "profile_function(_, {e}, {a})", e=event, a=arg)
+    trace_object(QUITE_VERBOSE, frame, "frame")
 
     # Resolve the names for the function (callable) and module
     name = "???"
@@ -653,23 +803,23 @@ def profile_function(frame, event, arg):
             method = getattr(frame.f_locals["self"], name, None)
             module = method.__module__ if method else None
         else:
-            trace(7, "Unable to resolve module")
+            trace(QUITE_VERBOSE, "Unable to resolve module")
     except:
         _print_exception_info("profile_function")
 
     # Trace out the call (with other information at higher tracing levels)
     if event.endswith("call"):
-        trace_fmt(4, "in: {mod}:{func}({a})", mod=module, func=name, a=arg)
+        trace_fmt(DETAILED, "in: {mod}:{func}({a})", mod=module, func=name, a=arg)
     elif event.endswith("return"):
-        trace_fmt(4, "out: {mod}:{func} => {a}", mod=module, func=name, a=arg)
+        trace_fmt(DETAILED, "out: {mod}:{func} => {a}", mod=module, func=name, a=arg)
     else:
-        trace_fmt(5, "profile {mod}:{func} {e}: arg={a}",
+        trace_fmt(VERBOSE, "profile {mod}:{func} {e}: arg={a}",
                   mod=module, func=name, e=event, a=arg)
     return
 
-def reference_vars(*args):
-    """Dummy function used for referencing variables"""
-    trace(9, f"reference_vars{tuple(args)}")
+def reference_var(*args):
+    """No-op function used for referencing variables in ARGS"""
+    trace(MOST_VERBOSE, f"reference_var{tuple(args)}")
     return
 
 #-------------------------------------------------------------------------------
@@ -700,26 +850,28 @@ def read_line(filename, line_number):
 #-------------------------------------------------------------------------------
 
 def main(args):
-    """Supporting code for command-line processing"""
-    trace_expr(4, len(args))
-    trace(1, "Warning: Not intended for direct invocation. A simple tracing example follows.")
+    """Supporting/ code for command-line processing"""
+    trace_expr(DETAILED, len(args))
+    trace(ERROR, "FYI: Not intended for direct invocation. Some tracing examples follow.")
     #
-    trace(0, "date record for now at trace level 1")
-    trace_object(1, datetime.now(), label="now")
+    trace(ALWAYS, "date record for now at trace level 1")
+    trace_object(ERROR, datetime.now(), label="now")
+    trace(DETAILED, "stack record with max depth 1")
+    trace_object(DETAILED, inspect.stack(), label="stack", max_depth=1)
     #
     # Make sure trace_expr traces at proper tracing level
-    trace(0, "level=N         for N from 0..trace_level ({l})".
+    trace(ALWAYS, "level=N         for N from 0..trace_level ({l})".
           format(l=trace_level))
     for level in range(trace_level):
         level_value = level
         trace_expr(level, level_value)
     #
     # Maker sure trace_expr gets all arguments
-    trace(0, "n-i=N-i         for N=trace_level ({l}) and i from 1..-1".format(l=trace_level))
+    trace(ALWAYS, "n-i=N-i         for N=trace_level ({l}) and i from 1..-1".format(l=trace_level))
     n = trace_level
     i = 1
     ## TODO: for i in range(3, 0, -1):
-    trace_expr(1, n+i, n, n+i)
+    trace_expr(ERROR, n+i, n, n+i)
     return
 
 # Do debug-only processing (n.b., for when PYTHONOPTIMIZE not set)
@@ -730,7 +882,7 @@ if __debug__:
     def debug_init():
         """Debug-only initialization"""
         time_start = time.time()
-        trace(4, "in debug_init()")
+        trace(DETAILED, "in debug_init()")
 
         # Open external file for copy of trace output
         global debug_file
@@ -747,15 +899,16 @@ if __debug__:
     
         # Show startup time and tracing info
         module_file = __file__
-        trace_fmtd(4, "[{f}] loaded at {t}", f=module_file, t=timestamp())
-        trace_fmtd(4, "trace_level={l}; output_timestamps={ots}", l=trace_level, ots=output_timestamps)
-        trace_fmtd(5, "debug_filename={fn} debug_file={f}",
+        trace_fmtd(DETAILED, "[{f}] loaded at {t}", f=module_file, t=timestamp())
+        trace_fmtd(DETAILED, "trace_level={l}; output_timestamps={ots}", l=trace_level, ots=output_timestamps)
+        trace_fmtd(VERBOSE, "debug_filename={fn} debug_file={f}",
                    fn=debug_filename, f=debug_file)
 
-        # Show additional information when detailed debugging
-        # TODO: sort keys to facilate comparisons of log files
-        trace_values(5, dict(os.environ), "environment")
-
+        # Determine other debug-only environment options
+        global para_mode_tracing
+        para_mode_tracing = _getenv_bool("PARA_MODE_TRACING", para_mode_tracing)
+        global max_trace_value_len
+        max_trace_value_len = _getenv_int("MAX_TRACE_VALUE_LEN", max_trace_value_len)
         global use_logging
         use_logging = _getenv_bool("USE_LOGGING", use_logging)
         enable_logging = _getenv_bool("ENABLE_LOGGING", use_logging)
@@ -764,17 +917,32 @@ if __debug__:
         monitor_functions = _getenv_bool("MONITOR_FUNCTIONS", False)
         if monitor_functions:
             sys.setprofile(profile_function)
+        trace_expr(DETAILED, para_mode_tracing, max_trace_value_len, use_logging, enable_logging, monitor_functions)
+
+        # Show additional information when detailed debugging
+        # TODO: sort keys to facilate comparisons of log files
+        pre = post = ""
+        if para_mode_tracing:
+            pre = post = "\n"
+        trace_fmt(DETAILED, "{pre}environment: {{\n\t{env}\n}}{post}",
+                  env="\n\t".join([(k + ': ' + os.environ[k]) for k in sorted(dict(os.environ))]),
+                  pre=pre, post=post)
+
+        # Likewise show additional information during verbose debug tracing
+        # Note: use debug.trace_current_context() in client module to show module-specific globals like __name__
+        trace_expr(VERBOSE, globals())
 
         # Register to show shuttdown time and elapsed seconds
         # TODO: rename to reflect generic-exit nature
         def display_ending_time_etc():
             """Display ending time information"""
             # note: does nothing if stderr closed (e.g., other monitor)
-            trace_object(7, sys.stderr)
-            if sys.stderr.closed:
+            # TODO: resolve pylint issue with sys.stderr.closed
+            trace_object(QUITE_VERBOSE, sys.stderr)
+            if sys.stderr.closed:       # pylint: disable=using-constant-test
                 return
             elapsed = round(time.time() - time_start, 3)
-            trace_fmtd(4, "[{f}] unloaded at {t}; elapsed={e}s",
+            trace_fmtd(DETAILED, "[{f}] unloaded at {t}; elapsed={e}s",
                        f=module_file, t=timestamp(), e=elapsed)
             if monitor_functions:
                 sys.setprofile(None)
@@ -796,4 +964,4 @@ if __debug__:
 if __name__ == '__main__':
     main(sys.argv)
 else:
-    trace_expr(9, 999)
+    trace_expr(MOST_VERBOSE, 999)
