@@ -3,6 +3,11 @@
 # Utility functions for writing glue scripts, such as implementing functionality
 # available in Unix scripting (e.g., basename command).
 #
+# NOTE:
+# - *** This module is being phased out with important functionality being moved elsewhere.
+# - Some of the utilities are specific to Unix (e.g., full_mkdir and real_path). In
+#   contrast, system.py attempts to be more cross platform.
+#
 # TODO:
 # - Add more functions to facilitate command-line scripting (check bash scripts for commonly used features).
 # - Add functions to facilitate functional programming (e.g., to simply debugging traces).
@@ -17,6 +22,7 @@ import inspect
 import os
 import re
 import shutil
+from subprocess import getoutput
 import sys
 import tempfile
 
@@ -25,21 +31,26 @@ import textwrap
 
 # Local packages
 from mezcla import debug
+from mezcla import system
 from mezcla import tpo_common as tpo
 from mezcla.tpo_common import debug_format, debug_print, print_stderr, setenv
 
-if sys.version_info[0] < 3:
-    from commands import getoutput       # pylint: disable=import-error
-else:
-    from subprocess import getoutput     # pylint: disable=no-name-in-module
-
-# note: ALLOW_SUBCOMMAND_TRACING should be interepreted in terms of detailed
+# note:
+# - ALLOW_SUBCOMMAND_TRACING should be interepreted in terms of detailed
 # tracing. Now, basic tracing is still done unless disable_subcommand_tracing()
-# invoked. (This way, subscript start/end time shown by default)
-ALLOW_SUBCOMMAND_TRACING = tpo.getenv_boolean("ALLOW_SUBCOMMAND_TRACING", False)
-default_subtrace_level = min(tpo.USUAL, tpo.debugging_level())
+# invoked. (This way, the subscript start/end time is still shown by default)
+# - SUB_DEBUG_LEVEL added to make sub-script trace level explicit
+DEFAULT_SUB_DEBUG_LEVEL = int(min(debug.TL.USUAL, debug.get_level()))
+SUB_DEBUG_LEVEL = system.getenv_int("SUB_DEBUG_LEVEL", DEFAULT_SUB_DEBUG_LEVEL,
+                                    "Tracing level for sub-command scripts invoked")
+default_subtrace_level = SUB_DEBUG_LEVEL
+ALLOW_SUBCOMMAND_TRACING = tpo.getenv_boolean("ALLOW_SUBCOMMAND_TRACING",
+                                              (SUB_DEBUG_LEVEL > DEFAULT_SUB_DEBUG_LEVEL),
+                                              "Whether sub-commands have tracing above TL.USUAL")
+## OLD: default_subtrace_level = min(tpo.USUAL, debug.get_level())
 if ALLOW_SUBCOMMAND_TRACING:
-    default_subtrace_level = tpo.debugging_level()
+    # TODO: work out intuitive default if both SUB_DEBUG_LEVEL and ALLOW_SUBCOMMAND_TRACING specified
+    default_subtrace_level = max(debug.get_level(), SUB_DEBUG_LEVEL)
 
 INDENT = "    "                          # default indentation
 
@@ -93,6 +104,21 @@ def remove_extension(filename, extension):
     return base
 
 
+## TODO: def replace_extension(filename, old_extension, old_extension):
+##           ...
+
+
+def dir_path(filename):
+    """Wrapper around os.path.dirname"""
+    # TODO: return . for filename without directory (not "")
+    # EX: dirname("/tmp/solr-4888.log") => "/tmp"
+    path = os.path.dirname(filename)
+    debug.trace(5, f"dirname({filename}) => {path}")
+    # TODO: add realpath (i.e., canonical path)
+    debug.assertion(form_path(path, basename(filename)) == filename)
+    return path
+
+
 def file_exists(filename):
     """Returns indication that FILENAME exists"""
     ok = os.path.exists(filename)
@@ -109,7 +135,8 @@ def non_empty_file(filename):
 
 
 def resolve_path(filename, base_dir=None):
-    """Resolves path for FILENAME related to BASE_DIR if not in current directory. Note: this uses the script directory for the calling module if BASE_DIR not specified (i.e., as if os.path.dirname(__file__) passed)."""
+    """Resolves path for FILENAME relative to BASE_DIR if not in current directory. Note: this uses the script directory for the calling module if BASE_DIR not specified (i.e., as if os.path.dirname(__file__) passed)."""
+    # TODO: give preference to script directory over current directory
     path = filename
     if not os.path.exists(path):
         if not base_dir:
@@ -161,6 +188,17 @@ def full_mkdir(path):
     return
 
 
+def real_path(path):
+    """Return resolved absolute pathname for PATH, as with Linux realpath command
+    Note: Use version in system instead"""
+    # EX: re.search("vmlinuz.*\d.\d", real_path("/vmlinuz"))
+    ## TODO: result = system.real_path(path)
+    result = run(f'realpath "{path}"')
+    debug.trace(6, "Warning: obsolete: use system.real_path instead")
+    debug.trace(7, f"real_path({path}) => {result}")
+    return result
+
+
 def indent(text, indentation=INDENT, max_width=512):
     """Indent TEXT with INDENTATION at beginning of each line, returning string ending in a newline unless empty and with resulting lines longer than max_width characters wrapped. Text is treated as a single paragraph."""
     # Note: an empty text is returned without trailing newline
@@ -186,10 +224,14 @@ def indent_lines(text, indentation=INDENT, max_width=512):
 
 MAX_ELIDED_TEXT_LEN = tpo.getenv_integer("MAX_ELIDED_TEXT_LEN", 128)
 #
-def elide(text, max_len=MAX_ELIDED_TEXT_LEN):
+def elide(text: str, max_len=None):
     """Returns TEXT elided to at most MAX_LEN characters (with '...' used to indicate remainder). Note: intended for tracing long string."""
+    # EX: elide("=" * 80, max_len=8) => "=====..."
     # TODO: add support for eliding at word-boundaries
+    tpo.debug_print("elide(_, _)", 7)
     debug.assertion(isinstance(text, str))
+    if max_len is None:
+        max_len = MAX_ELIDED_TEXT_LEN
     result = text
     if len(result) > max_len:
         result = result[:max_len] + "..."
@@ -197,15 +239,32 @@ def elide(text, max_len=MAX_ELIDED_TEXT_LEN):
     return result
 
 
+def elide_values(values: list, **kwargs):
+    """List version of elide [q.v.]"""
+    # EX: elide_values(["1", "22", "333"], max_len=2) => ["1", "22", "33..."]
+    tpo.debug_print("elide_values(_, _)", 7)
+    return list(map(lambda v: elide(str(v), **kwargs),
+                    values))
+
+
 def disable_subcommand_tracing():
-    """Disables tracing in scripts invoked via run().""" 
+    """Disables tracing in scripts invoked via run().
+    Note: Invoked in unittest_wrapper.py"""
+    tpo.debug_print("disable_subcommand_tracing()", 7)
     # Note this works by having run() temporarily setting DEBUG_LEVEL to 0."""
     global default_subtrace_level
     default_subtrace_level = 0
 
 
 def run(command, trace_level=4, subtrace_level=None, just_issue=False, **namespace):
-    """Invokes COMMAND via system shell, using TRACE_LEVEL for debugging output, returning result. The command can use format-style templates, resolved from caller's namespace. The optional SUBTRACE_LEVEL sets tracing for invoked commands (default is same as TRACE_LEVEL); this works around problem with stderr not being separated, which can be a problem when tracing unit tests. Notes: This function doesn't work fully under Win32. Tabs are not preserved, so redirect stdout to a file if needed."""
+    """Invokes COMMAND via system shell, using TRACE_LEVEL for debugging output, returning result. The command can use format-style templates, resolved from caller's namespace. The optional SUBTRACE_LEVEL sets tracing for invoked commands (default is same as TRACE_LEVEL); this works around problem with stderr not being separated, which can be a problem when tracing unit tests.
+   Notes:
+   - The result includes stderr, so direct if not desired:
+         gh.run("ls /tmp/fubar 2> /dev/null")
+   - This is only intended for running simple commands. It would be better to create a subprocess for any complex interactions.
+   - This function doesn't work fully under Win32. Tabs are not preserved, so redirect stdout to a file if needed.
+   """
+    # TODO: add automatic log file support as in run_script from unittest_wrapper.py
     # TODO: make sure no template markers left in command text (e.g., "tar cvfz {tar_file}")
     # EX: "root" in run("ls /")
     # Note: Script tracing controlled DEBUG_LEVEL environment variable.
@@ -232,6 +291,7 @@ def run(command, trace_level=4, subtrace_level=None, just_issue=False, **namespa
     # Note: Unix supports the '>|' pipe operator (i.e., output with overwrite); but,
     # it is not supported under Windows. To avoid unexpected porting issues, clients
     # should replace 'run("... >| f")' usages with 'delete_file(f); run(...)'.
+    # note: TestWrapper.setUp handles the deletion automatically
     assertion(">|" not in command_line)
     result = None
     ## TODO: if (just_issue or not wait): ... else: ...
@@ -249,7 +309,7 @@ def issue(command, trace_level=4, subtrace_level=None, **namespace):
     - Nothing is returned.
     - Traces stdout when debugging at quite-detailed level (6).
     - Captures stderr unless redirected and traces at error level (1)."""
-    # EX: issue("ls /") => ""
+    # EX: issue("ls /") => None
     # EX: issue("xeyes &")
     debug_print("issue(%s, [trace_level=%s], [subtrace_level=%s], [ns=%s])"
                 % (command, trace_level, subtrace_level, namespace), (trace_level + 1))
@@ -258,8 +318,7 @@ def issue(command, trace_level=4, subtrace_level=None, **namespace):
     if tpo.debugging() and (not "2>" in command) and (not "2|&1" in command):
         ## TODO: use a different suffix each time to aid in debugging
         log_file = TEMP_LOG_FILE
-        ## BAD: command += " 2>| " + log_file
-        delete_file(log_file)
+        delete_existing_file(log_file)
         command += " 2> " + log_file
     # Run the command and trace output
     command_line = command
@@ -280,36 +339,54 @@ def issue(command, trace_level=4, subtrace_level=None, **namespace):
 def get_hex_dump(text, break_newlines=False):
     """Get hex dump for TEXT, optionally BREAKing lines on NEWLINES"""
     # TODO: implement entirely within Pyton (e.g., via binascii.hexlify)
-    # EX: get_hex_dump
-    debug.trace_fmt(6, "get_hex_dump{{t}, {bn})", t=text, bn=break_newlines)
+    # EX: get_hex_dump("Tomás") => \
+    #   "00000000  54 6F 6D C3 A1 73       -                          Tom..s"
+    debug.trace_fmt(6, "get_hex_dump({t}, {bn})", t=text, bn=break_newlines)
     in_file = get_temp_file() + ".in.list"
     out_file = get_temp_file() + ".out.list"
-    write_file(in_file, text)
-    run("perl -Ss hexview.perl -newlines {i} > {o}", i=in_file, o=out_file)
-    result = read_file(out_file)    
+    ## BAD:
+    ## write_file(in_file, text)
+    ## run("perl -Ss hexview.perl -newlines {i} > {o}", i=in_file, o=out_file)
+    system.write_file(in_file, text, skip_newline=True)
+    run("perl -Ss hexview.perl {i} > {o}", i=in_file, o=out_file)
+    result = read_file(out_file).rstrip("\n")
+    debug.trace(7, f"get_hex_dump() => {result}")
     return result
 
 
-def extract_matches(pattern, lines, fields=1, multiple=False):
+def extract_matches(pattern, lines, fields=1, multiple=False, re_flags=0, para_mode=False):
     """Checks for PATTERN matches in LINES of text returning list of tuples with replacement groups.
     Notes: The number of FIELDS can be greater than 1.
-    Optionally allows for MULTIPLE matches within a line."""
+    Optionally allows for MULTIPLE matches within a line.
+    The lines are concatenated if DOTALL
+    Matching optionally uses perl-style PARA_MODE
+    """
+    ## TODO: If unspecified, the regex flags default to DOTALL.
     # ex: extract_matches(r"^(\S+) \S+", ["John D.", "Jane D.", "Plato"]) => ["John", "Jane"]
     # Note: modelled after extract_matches.perl
     # TODO: make multiple the default
-    debug_print("extract_matches(%s, _, [f=%s], [m=%s])" % (pattern, fields, multiple), 6)
+    debug_print("extract_matches(%s, _, [fld=%s], [m=%s], [flg=%s], [para=%s])" % (pattern, fields, multiple, re_flags, para_mode), 6)
+    ## TODO
+    ## if re_flags is None:
+    ##     re_flags = re.DOTALL
     debug.trace_values(6, lines, "lines")
     ## assert type(lines) == list
     assert isinstance(lines, list)
     if pattern.find("(") == -1:
         pattern = "(" + pattern + ")"
+    if (re_flags and (re_flags & re.DOTALL)):
+        lines = [ "\n".join(lines) + "\n" ]
+        debug.trace_expr(6, lines)
+    if para_mode:
+        lines = re.split(r"\n\s*\n", ("\n".join(lines)))
     matches = []
     for i, line in enumerate(lines):
         while line:
             debug.trace(6, f"L{i}: {line}")
             try:
                 # Extract match field(s)
-                match = re.search(pattern, line)
+                debug.trace_expr(7, pattern, line, re_flags, fields)
+                match = re.search(pattern, line, flags=re_flags)
                 if not match:
                     break
                 result = match.group(1) if (fields == 1) else [match.group(i + 1) for i in range(fields)]
@@ -325,6 +402,7 @@ def extract_matches(pattern, lines, fields=1, multiple=False):
                 line = new_line
             except (re.error, IndexError):
                 debug_print("Warning: Exception in pattern matching: %s" % str(sys.exc_info()), 2)
+                line = None
     debug_print("extract_matches() => %s" % (matches), 7)
     double_indent = INDENT + INDENT
     debug_format("{ind}input lines: {{\n{res}\n{ind}}}", 8,
@@ -332,21 +410,21 @@ def extract_matches(pattern, lines, fields=1, multiple=False):
     return matches
 
 
-def extract_match(pattern, lines, fields=1, multiple=False):
+def extract_match(pattern, lines, fields=1, multiple=False, re_flags=0, para_mode=None):
     """Extracts first match of PATTERN in LINES for FIELDS"""
-    matches = extract_matches(pattern, lines, fields, multiple)
+    matches = extract_matches(pattern, lines, fields, multiple, re_flags, para_mode)
     result = (matches[0] if matches else None)
     debug_print("match: %s" % result, 5)
     return result
 
 
-def extract_match_from_text(pattern, text, fields=1, multiple=False):
+def extract_match_from_text(pattern, text, fields=1, multiple=False, re_flags=0, para_mode=None):
     """Wrapper around extract_match for text input"""
     ## TODO: rework to allow for multiple-line matching
-    return extract_match(pattern, text.split("\n"), fields, multiple)
+    return extract_match(pattern, text.split("\n"), fields, multiple, re_flags, para_mode)
 
 
-def extract_matches_from_text(pattern, text, fields=1, multiple=None):
+def extract_matches_from_text(pattern, text, fields=1, multiple=None, re_flags=0, para_mode=None):
     """Wrapper around extract_matches for text input
     Note: By default MULTIPLE matches are returned"""
     # EX: extract_matches_from_text(".", "abc") => ["a", "b", "c"]
@@ -354,7 +432,7 @@ def extract_matches_from_text(pattern, text, fields=1, multiple=None):
     if multiple is None:
         multiple = True
     # TODO: make multiple True by default
-    return extract_matches(pattern, text.split("\n"), fields, multiple)
+    return extract_matches(pattern, text.split("\n"), fields, multiple, re_flags, para_mode)
 
 
 def count_it(pattern, text, field=1, multiple=None):
@@ -405,6 +483,7 @@ def write_lines(filename, text_lines, append=False):
     """Creates FILENAME using TEXT_LINES with newlines added and optionally for APPEND"""
     debug_print("write_lines(%s, _)" % (filename), 5)
     debug_print("    text_lines=%s" % text_lines, 6)
+    debug.assertion(isinstance(text_lines, list) and all(isinstance(x, str) for x in text_lines))
     f = None
     try:
         mode = 'a' if append else 'w'
