@@ -58,6 +58,9 @@
 # - Have option for processing text by page by page, instead of
 #   just tracking page number with
 # - Clarify the input processing in various modes: line, paragraph and file-input.
+# - Add function for getting temp_base as dir:
+#       dummy_app = Main(use_temp_base_dir=True)
+#       temp_wav_path = gh.form_path(dummy_app.temp_base, WAV_FILE)
 #
 
 """Module for encapsulating main() processing"""
@@ -65,6 +68,7 @@
 # Standard packages
 import argparse
 import os
+import re
 import sys
 import tempfile
 
@@ -77,17 +81,29 @@ from mezcla.system import getenv_bool
 
 # Constants
 HELP_ARG = "--help"
+USAGE_ARG = "--usage"
 USE_PARAGRAPH_MODE_DEFAULT = getenv_bool("USE_PARAGRAPH_MODE", False)
 USE_PARAGRAPH_MODE = getenv_bool("PARAGRAPH_MODE", USE_PARAGRAPH_MODE_DEFAULT,
                                  "Process input in Perl-style paragraph mode")
 FILE_INPUT_MODE = getenv_bool("FILE_INPUT_MODE", False,
                               "Read stdin using Perl-style file input mode--aka file slurping")
+debug.assertion(not (USE_PARAGRAPH_MODE and FILE_INPUT_MODE))
 
 FORM_FEED = "\f"
 TRACK_PAGES = getenv_bool("TRACK_PAGES", False,
                           "Track page boundaries by form feed--\\f or ^L")
 RETAIN_FORM_FEED = getenv_bool("RETAIN_FORM_FEED", False,
                                "Include formfeed (\\f) at start of each segment")
+DEFAULT_FILE_BASE = re.sub(r".py\w*$", "", gh.basename(__file__))
+FILE_BASE = system.getenv_text("FILE_BASE", DEFAULT_FILE_BASE,
+                               "Basename for output files including dir")
+SHOW_ENV_OPTIONS = system.getenv_bool("ENV_USAGE", debug.detailed_debugging(),
+                                      "Include environment options in usage")
+# TODO: Put following in common module
+INDENT = system.getenv_text("INDENT", "    ",
+                            "Indentation for system output")
+BRIEF_USAGE = system.getenv_bool("BRIEF_USAGE", False,
+                                 "Show brief usage with autohelp")
 
 #-------------------------------------------------------------------------------
 
@@ -96,29 +112,31 @@ class Main(object):
     argument_parser = None
     force_unicode = False
 
-    def __init__(self, runtime_args=None, description=None, 
+    def __init__(self, runtime_args=None, description=None, skip_args=False,
                  # TODO: Either rename xyz_optiom to match python type name 
                  # or rename them without abbreviations.
                  # TODO: explain difference between positional_options and positional_arguments
                  multiple_files=False,
                  use_temp_base_dir=None,
                  usage_notes=None,
+                 program=None,
                  paragraph_mode=None, track_pages=None, file_input_mode=None,
                  boolean_options=None, text_options=None, int_options=None,
                  float_options=None, positional_options=None, positional_arguments=None,
-                 skip_input=None, manual_input=None, auto_help=None):
+                 skip_input=None, manual_input=None, auto_help=None, brief_usage=None, **kwargs):
         """Class constructor: parses RUNTIME_ARGS (or command line), with specifications
         for BOOLEAN_OPTIONS, TEXT_OPTIONS, INT_OPTIONS, FLOAT_OPTIONS, and POSITIONAL_OPTIONS
         (see convert_option). Includes options to SKIP_INPUT, or to have MANUAL_INPUT, or to use AUTO_HELP invocation (i.e., assuming {ha} if no args)."""
-        tpo.debug_format("Main.__init__({args}, d={desc}, b={bools}, t={texts}, "
-                         + "i={ints}, f={floats}, p={posns}, s={skip}, m={mi}, a={auto},"
-                         + "pm={para}, tp={page}, fim={file}", 5,
+        tpo.debug_format("Main.__init__({args}, d={desc}, b={bools}, t={texts},"
+                         + " i={ints}, f={floats}, p={posns}, s={skip}, m={mi}, ah={auto}, bu={usage}"
+                         + " pm={para}, tp={page}, fim={file}, prog={prog}, noargs={skip_args}, kw={kw})", 5,
                          args=runtime_args, desc=description, bools=boolean_options,
                          texts=text_options, ints=int_options, floats=float_options,
-                         posns=positional_options, skip=skip_input, mi=manual_input, auto=auto_help,
+                         posns=positional_options, skip=skip_input, mi=manual_input,
+                         auto=auto_help, usage=brief_usage,
                          para=paragraph_mode, page=track_pages, file=file_input_mode,
-                         ha=HELP_ARG)
-        self.description = "TODO: what the script does" # defaults to TODO note for client
+                         prog=program, ha=HELP_ARG, skip_args=skip_args, kw=kwargs)
+        self.description = "TODO: what the script does"   # *** DONT'T MODIFY: defaults TODO note for client
         # TODO: boolean_options = [(VERBOSE, "Verbose output mode")]
         self.boolean_options = []
         self.text_options = []
@@ -152,10 +170,13 @@ class Main(object):
         self.skip_input = skip_input
         #
         self.parser = None
+        if brief_usage is None:
+            brief_usage = BRIEF_USAGE
+        self.brief_usage = brief_usage  # show brief usage instead of full --help
         if auto_help is None:
             ## OLD: auto_help = self.skip_input
             auto_help = self.skip_input or not self.manual_input
-        self.auto_help = auto_help
+        self.auto_help = auto_help      # adds --help to command line if no arguments
         if usage_notes is None:
             usage_notes = ""
         self.notes = usage_notes
@@ -171,10 +192,14 @@ class Main(object):
 
         # Setup temporary file and/or base directory
         # TODO: allow temp_base handling to be overridable by constructor options
+        prefix = (FILE_BASE + "-")
+        ntf_args = {'prefix': prefix,
+                    ## TODO: 'suffix': "-"
+                    }
         self.temp_base = tpo.getenv_text("TEMP_BASE",
-                                         tempfile.NamedTemporaryFile().name)
+                                         tempfile.NamedTemporaryFile(**ntf_args).name)
         # TODO: self.use_temp_base_dir = gh.dir_exists(gh.basename(self.temp_base))
-        # -or-: temp_base_dir = tpo.getenv_text("TEMP_BASE_DIR", ""); self.use_temp_base_dir = bool(temp_base_dir.strip); ...
+        # -or-: temp_base_dir = tpo.getenv_text("TEMP_BASE_DIR", ""); self.use_temp_base_dir = bool(temp_base_dir.strip()); ...
         if use_temp_base_dir is None:
             use_temp_base_dir = tpo.getenv_bool("USE_TEMP_BASE_DIR", False)
         self.use_temp_base_dir = use_temp_base_dir
@@ -188,13 +213,15 @@ class Main(object):
         # Get arguments from specified parameter or via command line
         # Note: --help assumed for input-less scripts with command line options
         # to avoid inadvertent script processing.
-        if runtime_args is None:
+        if ((runtime_args is None) and (not skip_args)):
             runtime_args = sys.argv[1:]
             tpo.debug_print("Using sys.argv[1:] for runtime args: %s" % runtime_args, 4)
             if self.auto_help and not runtime_args:
-                debug.trace_fmt(4, "Adding {ha} to command line (as per auto_help)", ha=HELP_ARG)
-                runtime_args = [HELP_ARG]
+                help_arg = (USAGE_ARG if self.brief_usage else HELP_ARG)
+                debug.trace(4, f"Adding {help_arg} to command line (as per auto_help)")
+                runtime_args = [help_arg]
         # Get other options
+        self.program = program
         if description:
             self.description = description
         if boolean_options:
@@ -209,13 +236,15 @@ class Main(object):
             # TODO: mark positional_options as decprecated
             debug.assertion(not (positional_options and positional_arguments))
             self.positional_options = positional_options or positional_arguments
-        self.multiple_files = multiple_files
+        self.multiple_files = multiple_files      # sets other_filenames if multiple w/ nargs=+ 
         # Set defaults
         self.parsed_args = None
         self.filename = None
         self.other_filenames = []
         # Do command-line parsing
-        self.check_arguments(runtime_args)
+        # TODO: consolidate with runtime_args check above
+        if not skip_args:
+            self.check_arguments(runtime_args)
         debug.trace_current_context(level=debug.QUITE_DETAILED)
         debug.trace_object(6, self, label="Main instance")
         debug.trace_fmt(tpo.QUITE_DETAILED, "end of Main.__init__(); self={s}",
@@ -227,10 +256,13 @@ class Main(object):
         Notes: The description and default of the specification are optional,
         and the parentheses can be omitted if just the label is given. Also,
         if POSITIONAL the option prefix (--) is omitted."""
+        ## TEST: result = ["", "", ""]
         opt_label = None
         opt_desc = None
         opt_default = default_value
+        opt_nargs = None
         opt_prefix = "--" if not positional else ""
+        # TODO: use keyword arguments (or namedtuple)
         if isinstance(option_spec, tuple):
             option_components = list(option_spec)
             opt_label = opt_prefix + option_components[0]
@@ -238,15 +270,27 @@ class Main(object):
                 opt_desc = option_components[1]
             if len(option_components) > 2:
                 opt_default = option_components[2]
+            if len(option_components) > 3:
+                opt_nargs = option_components[3]
+                debug.assertion(positional)
         else:
             opt_label = opt_prefix + tpo.to_string(option_spec)
         debug.assertion(not " " in opt_label)
-        result = (opt_label, opt_desc, opt_default)
+        result_list = [opt_label, opt_desc, opt_default]
+        if positional:
+            result_list.append(opt_nargs)
+        result = tuple(result_list)
         tpo.debug_format("convert_option({o}, {d}, {p}): self={s} => {r}", 5,
                          o=option_spec, d=default_value, p=positional,
                          s=self, r=result)
         return result
 
+    def convert_argument(self, argument_spec, default_value=None):
+        """Convert ARGUMENT_SPEC to (label, description, default) tuple. 
+        Note: This is a wrapper around convert_option for positional arguments."""
+        debug.trace(6, f"convert_option({argument_spec}, {default_value}")
+        return self.convert_option(argument_spec, default_value, positional=True)
+
     def get_option_name(self, label):
         """Return internal name for parser options (e.g. dashes converted to underscores)"""
         name = label.replace("-", "_")
@@ -254,18 +298,41 @@ class Main(object):
                          l=label, n=name, s=self)
         return name
 
-    def has_parsed_option(self, label):
-        """Whether option for LABEL specified (i.e., non-null value)"""
+    def has_parsed_option_old(self, label):
+        """Whether option for LABEL specified (i.e., non-null value)
+        Note: OLD version that checks for non-null value)
+        """
+        # EX: self.parsed_args = {'it': False}; self.has_parsed_option_old('nonit') => False
         name = self.get_option_name(label)
-        has_option = (name in self.parsed_args and self.parsed_args[name])
-        tpo.debug_format("has_parsed_option({l}) => {r}", 6,
+        has_option = (name in self.parsed_args and (self.parsed_args[name] is not None))
+        tpo.debug_format("has_parsed_option_old({l}) => {r}", 6,
                          l=label, r=has_option)
         return has_option
 
+    def has_parsed_option(self, label):
+        """Value for LABEL specified or None if not applicable
+        Note: This is a deprecated method (use get_parsed_option instead)
+        """
+        # EX: self.parsed_args = {'it': False}; self.has_parsed_option('notit') => None
+        ## TEMP HACK: if called by a subclass, treate as alias to get_parsed_option
+        if (self.__class__ != "__main__.Script"):
+            debug.trace(4, "Warning: deprecated method: has_parsed_option => get_parsed_option")
+            return self.get_parsed_option(label)
+        # Return parsed-arg entry for the option
+        name = self.get_option_name(label)
+        ## TEMP HACK: has_parsed_option returns args.get(opt)
+        option_value = self.parsed_args.get(name)
+        tpo.debug_format("has_parsed_option({l}) => {r}", 6,
+                         l=label, r=option_value)
+        return option_value
+
     def get_parsed_option(self, label, default=None, positional=False):
         """Get value for option LABEL, with dashes converted to underscores. 
         If POSITIONAL specified, DEFAULT value is used if omitted"""
         opt_label = self.get_option_name(label) if not positional else label
+        if not self.parsed_args:
+            debug.trace(5, "Error: Unexpected condition in get_parsed_option")
+            return default
         value = self.parsed_args.get(opt_label)
         # Override null value with default
         if value is None:
@@ -289,42 +356,45 @@ class Main(object):
         return self.get_parsed_option(label, default, positional=True)
 
     def check_arguments(self, runtime_args):
-        """Check command-line arguments"""
+        """Check command-line arguments
+        Note: This uses argparse, which might exit the process
+        """
         # Note: Shows env. options when debugging as these are backdoor settings.
         tpo.debug_format("Main.check_arguments({args})", 5, args=runtime_args)
         # TODO: add in detailed usage notes w/ environment option descriptions (see google_word2vec.py)
         if not self.argument_parser:
             self.argument_parser = argparse.ArgumentParser
         usage_notes = self.notes
-        if (not usage_notes and debug.debugging()):
-            env_opts = system.formatted_environment_option_descriptions(sort=True)
+        if (not usage_notes and SHOW_ENV_OPTIONS):
+            env_opts = system.formatted_environment_option_descriptions(sort=True, indent=INDENT)
             usage_notes = ("Notes: \n"
                            + ("- Use - for stdin to skip usage\n" if (not self.skip_input) else "")
-                           + ("- Available env. options:\n\t{opts}".format(
-                               opts=env_opts)))
+                           + ("- Available env. options:\n{indent}{opts}".format(
+                               opts=env_opts, indent=INDENT)))
         parser = self.argument_parser(description=self.description,
-                                      epilog=usage_notes,
+                                      epilog=usage_notes, prog=self.program,
                                       formatter_class=argparse.RawDescriptionHelpFormatter)
         # TODO: use capitalized script description but lowercase argument help
 
         # Check for options of specific types
-        # TODO: consolidate processing for the groups; add option for environment-based default
+        # TODO: consolidate processing for the groups; add option for environment-based default; resolve stupid pylint false positive about unbalanced-tuple-unpacking
         for opt_spec in self.boolean_options:
-            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)
+            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
             parser.add_argument(opt_label, default=opt_default, action='store_true',
                                 help=opt_desc)
         for opt_spec in self.int_options:
-            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)
+            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
             parser.add_argument(opt_label, type=int, default=opt_default, help=opt_desc)
         for opt_spec in self.float_options:
-            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)
+            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
             parser.add_argument(opt_label, type=float, default=opt_default,
                                 help=opt_desc)
         for opt_spec in self.text_options:
-            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)
+            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, None)    # pylint: disable=unbalanced-tuple-unpacking
             parser.add_argument(opt_label, default=opt_default, help=opt_desc)
 
         # Add dummy arguments
+        # Note: These are used as reminders on how to flesh out the initialization
         if tpo.detailed_debugging():
             if not self.boolean_options:
                 parser.add_argument("--TODO-bool-arg", default=False, action='store_true',
@@ -341,36 +411,45 @@ class Main(object):
 
         # Add positional arguments
         for i, opt_spec in enumerate(self.positional_options):
-            (opt_label, opt_desc, opt_default) = self.convert_option(opt_spec, "",
-                                                                     positional=True)
-            # note: a numeric nargs produces a list even if 1, so None used
-            nargs = None
+            (opt_label, opt_desc, opt_default, opt_nargs) = self.convert_argument(opt_spec, "")   # pylint: disable=unbalanced-tuple-unpacking
+            # note: a numeric nargs produces a list even if 1, so None used by default
+            nargs = opt_nargs
             tpo.debug_format("positional arg {i}, nargs={nargs}", 6, 
                              i=i, nargs=nargs)
+            # TODO: add default to opt_desc if not mentioned
             parser.add_argument(opt_label, default=opt_default, nargs=nargs, 
                                 help=opt_desc)
 
         # Add filename last and make optional with '-' default (stdin)
-        # Note: with nargs=+, the reult is a list of filename (even if one file [WTH?]!)
+        # Note: with nargs=+, the result is a list of filenames (even if one file [WTH?]!)
         if not self.skip_input:
             filename_nargs = ('?' if (not self.multiple_files) else "+")
             tpo.debug_format("filename_nargs={nargs}", 6, nargs=filename_nargs)
             parser.add_argument("filename", nargs=filename_nargs, default='-',
                                 help="Input filename")
 
-        # Parse the commandline and get result
+        # Optionally, show brief usage and exit
+        # note: print_usage just print command line synopsis (not individual descriptions)
+        if (self.brief_usage and (runtime_args == [USAGE_ARG])):
+            debug.trace(4, "Just showing (brief) usage and then exiting")
+            parser.print_usage()
+            sys.exit()
+
+        # Parse the command line and get result
         tpo.debug_format("parser={p}", 6, p=parser)
         self.parser = parser
         self.parsed_args = vars(parser.parse_args(runtime_args))
         tpo.debug_print("parsed_args = %s" % self.parsed_args, 5)
 
         # Get filename unless input ignored and fixup if returned as list
+        # TODO: add an option to retain self.filename as is
         if not self.skip_input:
             self.filename = self.parsed_args['filename']
             if (isinstance(self.filename, list)):
-                debug.trace(5, "Making (list) filename a string & other_files remainder")
+                if not self.multiple_files:
+                    debug.trace(3, "Warning: Making (list) self.filename a string & setting self.other_filenames to remainder")
                 file_list = self.filename
-                self.other_files = file_list[1:]
+                self.other_filenames = file_list[1:]
                 self.filename =  file_list[0] if len(file_list) else "-"
         debug.trace(6, "end Main.check_arguments()")
         return
@@ -401,23 +480,29 @@ class Main(object):
         return
 
     def run(self):
-        """Entry point for script"""
+        """Runner for script processing"""
         tpo.debug_print("Main.run()", 5)
         # TODO: decompose (e.g., isolate input proecessing)
 
-        # Have client do pre-input initialization
+        # Have client do pre-input initialization (e.g., argument extraction)
         self.setup()
 
         # Resolve input stream from either explicit filename or via standard input
         self.input_stream = sys.stdin
         if (self.filename and (self.filename != "-")):
-            debug.assertion(isinstance(self.filename, str))
-            debug.assertion(self.filename != ["-"])
-            debug.assertion(os.path.exists(self.filename))
-            self.input_stream = system.open_file(self.filename)
-            debug.assertion(self.input_stream)
+            if (isinstance(self.filename, list) or (len(self.other_filenames) > 0)):
+                debug.assertion(self.filename != ["-"])
+                if not self.multiple_files:
+                    # note: check_arguments sets self.other_filenames
+                    debug.trace(3, "Warning: Not opening multiple-valued filename arg")
+                    debug.trace_expr(3, self.filename, self.other_filenames)
+            else:
+                debug.assertion(isinstance(self.filename, str))
+                debug.assertion(os.path.exists(self.filename))
+                self.input_stream = system.open_file(self.filename)
+                debug.assertion(self.input_stream)
     
-        # Invoke the script
+        # Initiate the main input processing
         try:
             # If not automatic input, process the main step of script
             if self.manual_input:
@@ -617,9 +702,21 @@ class Main(object):
                 gh.run("rm -vf {file}*", file=self.temp_file)
         return
 
+#-------------------------------------------------------------------------------
+# Global instance for convenient adhoc usage
+# 
+# note: useful fo temporary file support (e.g., dummy_app.temp_file)
+#
+
+dummy_app = Main([])
+
+debug.trace_current_context(8, "main.py context")
+
 #------------------------------------------------------------------------
 
 if __name__ == '__main__':
     system.print_stderr(f"Warning: {__file__} is not intended to be run standalone")
     main = Main()
+    ## TODO??: main.description = "Script class support"
     main.run()
+
