@@ -19,9 +19,10 @@ import sys
 import more_itertools
 
 # Local packages
-import mezcla.debug as debug
-import mezcla.system as system
-import mezcla.text_utils as text_utils
+from mezcla import debug
+from mezcla import glue_helpers as gh
+from mezcla import system
+from mezcla import text_utils
 
 # Constants
 ELLIPSIS = "\u2026"                 # Horizontal Ellipsis
@@ -212,21 +213,45 @@ def string_diff(text1, text2):
     # EX: string_diff("one\ntwo\nthree\nfour", "one\ntoo\ntree\nfour") => "  one\n- two\n…  ^\n+ too\n…  ^\n- three\n…  -\n+ tree  four\n"
     
     # Perform comparison
-    lines1 = (text1 + "\n").splitlines(keepends=True)
-    lines2 = (text2 + "\n").splitlines(keepends=True)
-    diff_result = "".join(ndiff(lines1, lines2))
+    diff_result = "n/a"
+    try:
+        lines1 = (text1 + "\n").splitlines(keepends=True)
+        lines2 = (text2 + "\n").splitlines(keepends=True)
+        diff_result = "".join(ndiff(lines1, lines2))
+    except:
+        system.print_exception_info("string_diff compare")
     
     # Convert to diff-style output, using < and > instead of - and +
-    ## TODO:
-    ## diff_result = re.sub(r"^- ", "< ", diff_result, re.MULTILINE)
-    ## diff_result = re.sub(r"^\+ ", "> ", diff_result, re.MULTILINE)
-    ## diff_result = re.sub(r"^\? ", f"{ELLIPSIS} ", diff_result, re.MULTILINE)
-    diff_result = diff_result.replace("- ", "< ", 1).replace("\n- ", "\n< ").   \
-                  replace("+ ", "> ", 1).replace("\n+ ", "\n> ").               \
-                  replace("\n? ", f"\n{ELLIPSIS} ")
+    try:
+        ## TODO:
+        ## diff_result = re.sub(r"^- ", "< ", diff_result, re.MULTILINE)
+        ## diff_result = re.sub(r"^\+ ", "> ", diff_result, re.MULTILINE)
+        ## diff_result = re.sub(r"^\? ", f"{ELLIPSIS} ", diff_result, re.MULTILINE)
+        diff_result = diff_result.replace("- ", "< ", 1).replace("\n- ", "\n< ").   \
+                          replace("+ ", "> ", 1).replace("\n+ ", "\n> ").           \
+                          replace("\n? ", f"\n{ELLIPSIS} ")
+    except:
+        system.print_exception_info("string_diff postprocess")
+
     debug.trace(6, f"string_diff() => {diff_result}")
     return diff_result
 
+
+def elide_string_values(obj, depth=0, max_len=None):
+    """Elide the values of all strings in OBJ, which can either be a scalar, a list or a hash"""
+    if max_len is None:
+        max_len = 32
+    MAX_DEPTH = 10
+    if isinstance(obj, list) and (depth < MAX_DEPTH):
+        obj = [elide_string_values(v, depth=(1 + depth), max_len=max_len) for v in obj]
+    elif isinstance(obj, dict) and (depth < MAX_DEPTH):
+        obj = {k: elide_string_values(obj[k], depth=(1 + depth), max_len=max_len) for k in obj.keys()}
+    elif isinstance(obj, str):
+        obj = gh.elide(obj, max_len=max_len)
+    else:
+        pass
+    return obj
+
 #-------------------------------------------------------------------------------
 
 def main(args):
