@@ -14,12 +14,21 @@
 # - Also does k-fold cross validation over development data split using 1/k-th as test.
 # - That is, the training data is partitioned twice.
 # - As an expediency to disable validation, epsilon is used for validation percent (e.g., 1e-6), because sklearn doesn't allow the percent to be specified as zero.
+# - Keep changes in sync with pandas_sklearn.py (e.g., XGBoost and GPU options).
+#................................................................................
+# Debugging usage:
+#
+# module=pandas_sklearn
+# let PSL++; dataset_path=examples/iris.csv; dataset_name=$(basename "$dataset_path" .csv); log="_$module.$dataset_name.$(TODAY).$PSL.log"; DEBUG_LEVEL=5 OUTPUT_CSV=1 XGB_SKIP_GPU=1 CLASSIFIER=XGB COERCE_FLOAT=1 ENCODE_CLASSES=1 VERBOSE=1 python -m mezcla.$module $dataset > "$log" 2>&1
+#
+#................................................................................
 #
 # TODO:
 # - Add exception handling throughout.
 # - Replace iris.csv with non-trivial ML data (e.g., to serve in benchmarking).
 # - Make sure validation used in same tense as standard practice.
 # - Rename (e.g., pandas_sklearn.py => evaluate_classifiers.py).
+# - Decompose main() into more helpers.
 # 
 
 """Illustrates sklearn classification over data with panda csv-based import"""
@@ -28,7 +37,7 @@
 from mezcla import debug            # pylint: disable=ungrouped-imports
 
 # Standard packages
-import sys
+## OLD: import sys
 
 # Installed pckages
 import numpy as np
@@ -58,6 +67,9 @@ from mezcla.text_utils import getenv_ints
 
 #................................................................................
 # Constants (e.g., environment-based options)
+#
+# Note: Some classifiers are not part of sklearn and thus require additional support
+# to enable. These include XBG, Keras, and Auto-Sklearn.
 
 VERBOSE = getenv_bool("VERBOSE", False)
 DATA_FILE = getenv_text("DATA_FILE", "tests/iris.csv")
@@ -65,7 +77,8 @@ FIELD_SEP = getenv_text("FIELD_SEP", ",",
                         description="Field separator (delimiter)")
 SKIP_DATAFRAME = getenv_bool("SKIP_DATAFRAME", False)
 USE_DATAFRAME = getenv_bool("USE_DATAFRAME", not SKIP_DATAFRAME)
-IGNORE_FIELDS = text_utils.extract_string_list(getenv_text("IGNORE_FIELDS", ""))
+IGNORE_FIELDS = text_utils.extract_string_list(
+    getenv_value("IGNORE_FIELDS") or "")
 INCLUDE_PLOTS = getenv_bool("INCLUDE_PLOTS", False)
 DATASET_PLOTS = getenv_bool("DATASET_PLOTS", INCLUDE_PLOTS,
                             "Include plots summarizing dataset")
@@ -75,7 +88,9 @@ SKIP_VALIDATION = getenv_bool("SKIP_VALIDATION", False)
 KERAS_CLASSIFIER = "Keras"
 KNN_CLASSIFIER = "KNN"
 XGB_CLASSIFIER = "XGB"
-DEFAULT_CLASSIFIER = getenv_text("CLASSIFIER", "LR")
+LR_CLASSIFIER = "LR"
+DEFAULT_CLASSIFIER = getenv_text("CLASSIFIER", LR_CLASSIFIER,
+                                 "Label for default classifier")
 DEFAULT_DEVEL_CLASSIFIER = DEFAULT_CLASSIFIER if (not SKIP_VALIDATION) else None
 DEVEL_CLASSIFIER = getenv_text("DEVEL_CLASSIFIER", DEFAULT_DEVEL_CLASSIFIER)
 INCLUDE_ALL = getenv_bool("INCLUDE_ALL", False,
@@ -94,9 +109,9 @@ VALIDATION_CLASSIFIER = getenv_text("VALIDATION_CLASSIFIER", DEFAULT_VALIDATION_
 ## TEST: debug.assertion(False)
 SKIP_DEVEL = getenv_bool("SKIP_DEVEL", False)
 TEST_PCT = getenv_number("TEST_PCT", 0.10)
-SEED = getenv_bool("SEED", 7919)
+SEED = getenv_int("SEED", 7919)
 SCORING_METRIC = getenv_text("SCORING_METRIC", "accuracy")
-CLASS_VAR = getenv_text("CLASS_VAR", "")
+CLASS_VAR = getenv_value("CLASS_VAR", "")
 ## TEST: 
 ## SKIP_XGB = getenv_bool("SKIP_XGB", False)
 ## USE_XGB = getenv_bool("USE_XGB", not SKIP_XGB)
@@ -118,21 +133,35 @@ NUMERIC_CLASSES = getenv_bool("NUMERIC_CLASSES", False,
                               "Encode classes as number")
 #
 DEFAULT_INCLUDE_KERAS_CLASSIFIER = KERAS_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER]
-INCLUDE_KERAS = getenv_bool("INCLUDE_KERAS", DEFAULT_INCLUDE_KERAS_CLASSIFIER or INCLUDE_ALL)
+INCLUDE_KERAS = getenv_bool("INCLUDE_KERAS", DEFAULT_INCLUDE_KERAS_CLASSIFIER or INCLUDE_ALL,
+                            "Include Keras deep learning classifier")
 SKIP_KERAS = getenv_bool("SKIP_KERAS", not INCLUDE_KERAS)
 HIDDEN_UNIT_VALUES = getenv_ints("HIDDEN_UNIT_VALUES", "50, 100, 50")
 NUM_EPOCHS = getenv_int("NUM_EPOCHS", 100)
 BATCH_SIZE = getenv_int("BATCH_SIZE", None)
-
-GPU_DEVICE = system.getenv_text("GPU_DEVICE", "",
-                                "Device number for GPU (e.g., shown under nvidia-smi)")
+#
+AUTOSKLEARN_CLASSIFIER = "Auto-Sklearn"
+DEFAULT_INCLUDE_AUTOSKLEARN_CLASSIFIER = AUTOSKLEARN_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER]
+INCLUDE_AUTOSKLEARN = getenv_bool("INCLUDE_AUTOSKLEARN", DEFAULT_INCLUDE_AUTOSKLEARN_CLASSIFIER,
+                                  "Include Auto-Sklearn classifier for AutoML")
+#
+GPU_DEVICE = getenv_value("GPU_DEVICE", "",
+                          "Device number for GPU (e.g., shown under nvidia-smi)")
 SHOW_ABLATION = getenv_bool("SHOW_ABLATION", False,
                             "Show ablation plot for accuracy")
 PRECISION_RECALL = getenv_bool("PRECISION_RECALL", False,
                                "Plot precision/recall curve")
+MICRO_AVERAGE = getenv_bool("MICRO_AVERAGE", False,
+                            "Use micro-averaging for mutliclass problems")
+DUMP_MODEL = getenv_bool("DUMP_MODEL", False,
+                         "Dump out model-specific representation")
+
+# Globals
+devel_classifiers = [DEVEL_CLASSIFIER]
+validation_classifiers = [VALIDATION_CLASSIFIER]
 
 #...............................................................................
-# Optional packahes
+# Optional packages
 
 if INCLUDE_PLOTS:
     # pylint: disable=import-outside-toplevel, import-error
@@ -141,18 +170,28 @@ if INCLUDE_PLOTS:
 if INCLUDE_XGB:
     # pylint: disable=import-outside-toplevel, import-error
     import xgboost as xgb
-    if getenv_text("USE_XGB"):
-        system.print_stderr("Warning: deprecated option USE_XGB")
+    devel_classifiers.append(XGB_CLASSIFIER)
+    validation_classifiers.append(XGB_CLASSIFIER)
+if getenv_value("USE_XGB"):
+    system.print_stderr("Warning: deprecated option USE_XGB")
 ## Mote: Following added for tracking down segmentation fault
 ## DEBUG: print("after xgboost import")
     
 if INCLUDE_KERAS:
     # pylint: disable=import-outside-toplevel, ungrouped-imports, import-error
     from mezcla.keras_param_search import MyKerasClassifier, create_keras_model
+    devel_classifiers.append(KERAS_CLASSIFIER)
+    validation_classifiers.append(KERAS_CLASSIFIER)
 ## DEBUG: print("after keras_param_search import")
 
-if (INCLUDE_ALL and (not (INCLUDE_XGB or INCLUDE_KERAS))):
-    debug.trace(4, "Warning: Need to specify INCLUDE_(XGB/KERAS) separately")
+if INCLUDE_AUTOSKLEARN:
+    # pylint: disable=import-outside-toplevel, import-error
+    import autosklearn.classification
+    devel_classifiers.append(AUTOSKLEARN_CLASSIFIER)
+    validation_classifiers.append(AUTOSKLEARN_CLASSIFIER)
+
+if (INCLUDE_ALL and (not (INCLUDE_XGB or INCLUDE_KERAS or INCLUDE_AUTOSKLEARN))):
+    debug.trace(4, "Warning: Need to specify INCLUDE_(XGB/KERAS/AUTOSKLEARN) separately")
 
 #...............................................................................
 # Utility functions
@@ -183,16 +222,22 @@ def main():
     # TODO: convert into using main.py's Main class
     data_file = None
     args = system.get_args()
-    if ((len(args) > 1) and (args[1] == "--help")):
+    show_help = ((len(args) > 1) and (args[1] == "--help"))
+    show_usage = ((len(args) > 1) and (args[1] == "--usage"))
+    
+    if ((len(args) <= 1) or show_help or show_usage):
         script = gh.basename(args[0])
-        system.print_stderr("Usage: {scr} [--help] [data-file]", scr=script)
-        system.print_stderr("")
-        system.print_stderr("where data-file is in CSV format")
+        system.print_stderr("Usage: {scr} [--help | --usage] [data-file | -]", scr=script)
         system.print_stderr("")
+        system.print_stderr("Notes:")
+        system.print_stderr("- Use - for data-file to use default.")
+        system.print_stderr("- Where data-file is in CSV format (or TSV)")
         ## TODO: show_details = VERBOSE or debug.debugging()
-        if debug.debugging(2):
+        debugging = debug.detailed_debugging()
+        if (show_help or debugging):
             ## TODO: add legend
-            system.print_stderr("\t" + system.formatted_environment_option_descriptions(include_all=True))
+            system.print_stderr("- Environment options:")
+            system.print_stderr("\t" + system.formatted_environment_option_descriptions(include_all=debugging))
         system.exit("")
     if ((len(args) > 1) and (not args[1].startswith("-"))):
         data_file = args[1]
@@ -251,9 +296,7 @@ def main():
         print("Null count:")
         print(dataset.isnull().sum())
     except:
-        debug.trace_fmtd(2, "Error: Problem during dataset illustration, etc.: {exc}",
-                         exc=sys.exc_info())
-        debug.raise_exception(6)
+        system.print_exception_info("dataset illustration")
     
     # Optionally show some plot
     ## OLD: if (not SKIP_PLOTS):
@@ -291,17 +334,32 @@ def main():
         # TODO: NUMERIC_CLASSES
     if ENCODE_CLASSES:
         y_encodings = create_feature_mapping(y)
+        y_enum = sorted(y_encodings.keys(), key=lambda k: y_encodings[k])
+        debug.trace_expr(4, y_encodings, y_enum)
         y = np.array([y_encodings[v] for v in y])
+        if USE_DATAFRAME:
+            y = pd.DataFrame(y, columns=[class_var])
         # TODO: COERCE_FLOAT
     debug.trace_fmtd(7, "X={X}\ny={y}", X=X, y=y)
+    if USE_DATAFRAME:
+        if not isinstance(X, pd.DataFrame):
+            X = pd.DataFrame(y, columns=feature_names)
+        if not isinstance(y, pd.DataFrame):
+            y = pd.DataFrame(y, columns=[class_var])        
     if OUTPUT_CSV:
         # TODO: drop pandas index column (first one; no header)
         debug.assertion(USE_DATAFRAME)
-        debug.assertion(not (COERCE_FLOAT or NUMERIC_CLASSES or ENCODE_CLASSES))
+        # note: prevously X and Y aren't frames if coerced or otherwise converted
+        ## OLD:
+        ## debug.assertion(not (COERCE_FLOAT or NUMERIC_CLASSES or ENCODE_CLASSES))
+        debug.assertion((isinstance(X, pd.DataFrame) and isinstance(y, pd.DataFrame)))
         basename = system.remove_extension(data_file)
+        # TODO: combine into single dataframe and use to_csv over that; use .csv instead of .csv.list (and make sure not to overwrite)
         X.to_csv(basename + "-X.csv.list", sep=FIELD_SEP, index=False)
         y.to_csv(basename + "-y.csv.list", sep=FIELD_SEP, index=False)
-        debug.assertion(system.file_exists(basename + "-X.csv.list"))
+        gh.run("paste --delimiters='{d}' {b}-X.csv.list {b}-y.csv.list > {b}.csv.list",
+               d=FIELD_SEP, b=basename)
+        debug.assertion(system.file_exists(basename + ".csv.list"))
     X_train, X_validation, y_train, y_validation = model_selection.train_test_split(X, y, test_size=VALIDATION_PCT, random_state=SEED)
     ## TODO:
     debug.trace_fmtd(6, "X_train={xt}\nX_valid={xv}\ny_train={yt}\ny_valid={yv}", xt=X_train, xv=X_validation, yt=y_train, yv=y_validation)
@@ -338,7 +396,8 @@ def main():
         models.append((KERAS_CLASSIFIER,
                        MyKerasClassifier(hidden_units=HIDDEN_UNIT_VALUES, epochs=NUM_EPOCHS,
                                          batch_size=BATCH_SIZE, build_fn=create_model_fn)))
-    models.append(("LR", LogisticRegression()))
+    if (INCLUDE_ALL or (LR_CLASSIFIER in [DEFAULT_DEVEL_CLASSIFIER, DEFAULT_VALIDATION_CLASSIFIER])):
+        models.append((LR_CLASSIFIER, LogisticRegression()))
     if INCLUDE_MISC_CLASSIFIERS:
         models.append(("LDA", LinearDiscriminantAnalysis()))
         models.append((KNN_CLASSIFIER, KNeighborsClassifier()))
@@ -346,6 +405,11 @@ def main():
         models.append(("GNB", GaussianNB()))
         models.append(("MNB", MultinomialNB()))
         models.append(("SVM", SVC()))
+    if INCLUDE_AUTOSKLEARN:
+        models.append((AUTOSKLEARN_CLASSIFIER, autosklearn.classification.AutoSklearnClassifier()))
+    debug.trace_expr(5, models)
+    if not models:
+        system.exit("Error: no models defined")
 
     # Evaluate each model in turn.
     # TODO: show precision, recall, F1, as well as accuracy
@@ -363,12 +427,12 @@ def main():
         print("Sample development test set results using scoring method '{sm}'".format(sm=SCORING_METRIC))
         ## TODO: average = "micro" if (not is_binary) else None
         for name, model in models:
-            if ((name != DEVEL_CLASSIFIER) and (not INCLUDE_ALL_DEVEL)):
-                debug.trace_fmt(5, "Skipping classifier {n} (not devel and not include all)", n=name)
+            if ((name not in devel_classifiers) and (not INCLUDE_ALL_DEVEL)):
+                debug.trace_fmt(5, "Skipping classifier {n} (not for devel and not include all)", n=name)
                 continue
             kfold = model_selection.KFold(n_splits=10, shuffle=True, random_state=SEED)
             ## TODO: cv_results = model_selection.cross_val_score(model, X_train, y_train, cv=kfold, scoring=SCORING_METRIC, average=average)
-            ## TODO: get this to work when SCORING_METRIC is not accuracy (which leads to not supported error for multicalss data)
+            ## TODO: get this to work when SCORING_METRIC is not accuracy (which leads to not supported error for multiclass data)
             ## (e.g., add environment variable so that sklearn uses micro or macro average
             try:
                 cv_results = model_selection.cross_val_score(model, X_train, y_train, cv=kfold, scoring=SCORING_METRIC)
@@ -392,15 +456,11 @@ def main():
                         print("Feature importance:")
                         ## print(model.get_score())
                         print(sorted(zip(feature_names, model.feature_importances_),
-                                     ## TODO: key=(lambda name, score: score),
                                      key=lambda name_score: name_score[1],
                                      reverse=True))
                                           
             except:
-                debug.trace_fmtd(2, "Error: Problem during training evaluation: {exc}",
-                                 exc=sys.exc_info())
-                summaries.append("Exception for " + name)
-                debug.raise_exception(6)
+                system.print_exception_info("training evaluation")
         print("Cross validation results over development test set")
         print("name\tacc\tstdev")
         print("\n".join(summaries))
@@ -421,8 +481,8 @@ def main():
     average = "micro" if (not is_binary) else "binary"
     num_run = 0
     for name, model in models:
-        if ((name != VALIDATION_CLASSIFIER) and (not VALIDATE_ALL)):
-            debug.trace_fmt(5, "Skipping classifier {n} (not validation and not include all)", n=name)
+        if ((name not in validation_classifiers) and (not VALIDATE_ALL)):
+            debug.trace_fmt(5, "Skipping classifier {n} (not for validation and not include all)", n=name)
             continue
         try:
             ## DEBUG: debug.trace(5, f"X/Y_train types: {[type(v) for v in [X_train, y_train]]}")
@@ -437,11 +497,18 @@ def main():
                 print("training confusion matrix:")
                 training_predictions = model.predict(X_train)
                 print(confusion_matrix(y_train, training_predictions))
+            if DUMP_MODEL:
+                if (name == XGB_CLASSIFIER):
+                    basename = system.remove_extension(data_file)
+                    model.get_booster().dump_model(f"{basename}.model-dump.list",
+                                                   with_stats=True)
+                else:
+                    system.print_error(f"Warning: no model dump support for {name}")
 
             predictions = model.predict(X_validation)
             print("validation confusion matrix:")
             print(confusion_matrix(y_validation, predictions))
-            ## TODO: drop accurracy ... F1 (provide in report)
+            ## TODO: drop accuracy ... F1 (provide in report)
             if VERBOSE:
                 print("accuracy:", end=" ")
                 print(round_num(accuracy_score(y_validation, predictions)))
@@ -456,15 +523,14 @@ def main():
             print("classification report:")
             print(classification_report(y_validation, predictions))
             if PRECISION_RECALL:
-                show_precision_recall(name, model, X_train, y_train, X_validation, y_validation)
+                if MICRO_AVERAGE:
+                    show_average_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation)
+                else:
+                    show_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation)
             if SHOW_ABLATION:
                 show_ablation(name, model, X_train, y_train, X_validation, y_validation)
         except:
-            ## OLD:
-            ## debug.trace_fmtd(2, "Error: Problem during validation evaluation: {exc}",
-            ##                  exc=sys.exc_info())
-            debug.trace_exception(2, "validation evaluation")
-            debug.raise_exception(6)
+            system.print_exception_info("validation evaluation")
     if ((num_run == 0) and VALIDATION_CLASSIFIER):
         system.print_stderr("Error: Validation classifier '{clf}' not supported", clf=VALIDATION_CLASSIFIER)
 
@@ -494,17 +560,26 @@ def show_ablation(name, model, X_train, y_train, X_validation, y_validation):
     debug.trace(5, "end show_ablation")
     return
 
-def show_precision_recall(name, model, X_train, y_train, X_validation, y_validation):
-    """Show precision/recall results for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION"""
-    # based on https://www.statology.org/precision-recall-curve-python/
+def show_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation):
+    """Show precision/recall results for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION
+    Note: This is only for binary classification
+    """
+    # based on https://www.statology.org/precision-recall-curve-python
+    num_cases = len(y_validation)
+    debug.assertion(num_classes == 2)
     precision = recall = thresholds = []
-    debug.trace(4, f"show_precision_recall{tuple([name, model, X_train, y_train, X_validation, y_validation])}")
+    debug.trace(4, f"show_precision_recall{tuple([name, model, num_classes, X_train, y_train, X_validation, y_validation])}")
     try:
         model.fit(X_train, y_train)
         y_scores = model.predict_proba(X_validation)[:, 1]
         precision, recall, thresholds = precision_recall_curve(y_validation, y_scores)
     except:
         system.print_exception_info("show_precision_recall")
+    num_points = len(precision)
+    debug.trace_expr(5, num_cases, num_points, len(precision), precision, recall, thresholds)
+    ## TODO: debug.assertion(num_points <= (num_cases / 2))
+
+    # Show the plots or print the data
     if (not SKIP_PLOTS):
         plt.plot(recall, precision)
         ## TODO: plt.yticks(thresholds)
@@ -523,6 +598,65 @@ def show_precision_recall(name, model, X_train, y_train, X_validation, y_validat
     debug.trace(5, "end show_precision_recall")        
     return
     
+
+def show_average_precision_recall(name, model, num_classes, X_train, y_train, X_validation, y_validation):
+    """Show precision/recall results for X_TRAIN, Y_TRAIN, X_VALIDATION, Y_VALIDATION
+    Note: this is an approximation based on micro-averaging"""
+    # based loosely on combination of https://www.statology.org/precision-recall-curve-python
+    # and https://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html
+    debug.trace(4, f"show_average_precision_recall{tuple([name, model, num_classes, X_train, y_train, X_validation, y_validation])}")
+
+    # Compute the metrics for each class
+    num_cases = len(y_validation)
+    precision = [None] * num_classes
+    recall = [None] * num_classes
+    thresholds = [None] * num_classes
+    try:
+        model.fit(X_train, y_train)
+        for i in range(num_classes):
+            y_scores = model.predict_proba(X_validation)[:, i]
+            precision[i], recall[i], thresholds[i] = precision_recall_curve(y_validation, y_scores,
+                                                                            pos_label=i)
+    except:
+        system.print_exception_info("show_average_precision_recall compute")
+        
+    # Compute micro-average, quantifying score on all classes jointly
+    # TODO: do summation on previous loop
+    num_points = len(precision)
+    average_precision = [0] * num_points
+    average_recall = [0] * num_points
+    try:
+        for s in range(num_points):
+            for i in range(num_classes):
+                average_precision[s] += precision[i][s]
+                average_recall[s] += recall[i][s]
+            average_precision[s] /= num_classes
+            average_recall[s] /= num_classes
+    except:
+        system.print_exception_info("show_average_precision_recall micro")
+    debug.trace_expr(5, num_cases, num_points, len(precision), precision, recall, thresholds,
+                     average_precision, average_recall)
+    ## TODO: debug.assertion(num_points <= (num_cases / 2))
+
+    # Show the plots or print the data
+    if (not SKIP_PLOTS):
+        try:
+            plt.plot(average_recall, average_precision)
+            plt.xlabel("mean recall")
+            plt.ylabel("mean precision")
+            plt.title("precision recall curve [micro-averaged]")
+            plt.show()
+        except:
+            system.print_exception_info("show_average_precision_recall plot")
+    else:
+        print("precision:\n\t{vals!r}".format(vals=precision))
+        print("recall:\n\t{vals!r}".format(vals=recall))
+        print("thresholds:\n\t{vals!r}".format(vals=thresholds))
+        print("average precision:\n\t{vals!r}".format(vals=average_precision))
+        print("average recall:\n\t{vals!r}".format(vals=average_recall))
+    debug.trace(5, "end show_average_precision_recall")        
+    return
+
 #------------------------------------------------------------------------
 
 if __name__ == "__main__":
