@@ -1,18 +1,20 @@
 #! /usr/bin/env python
 #
-# TODO: Test(s) for ../MODULE.py
+# TODO: Test(s) for ../<module>.py
 #
 # Notes:
 # - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
 # - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
 #   (see examples below for setUp and tearDown).
+# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
+#   option shows tracing output normally suppressed by  unittest_wrapper.py.
 # - This can be run as follows:
-#   $ PYTHONPATH=".:$PYTHONPATH" python tests/test_MODULE.py
+#   $ PYTHONPATH="$(realpath .)/..):$PYTHONPATH" python tests/test_<module>.py
 #
 
-"""TODO: Tests for TODO:module module"""
+"""TODO: Tests for <module> module"""
 
-# Installed packages
+# Standard packages
 import re
 import unittest
 
@@ -21,19 +23,33 @@ import unittest
 
 # Local packages
 from mezcla.unittest_wrapper import TestWrapper
-import mezcla.glue_helpers as gh
-import mezcla.debug as debug
+from mezcla import debug
+## TODO: from mezcla import system
+from mezcla import glue_helpers as gh
 
 # Note: Two references are used for the module to be tested:
 #    THE_MODULE:	    global module object
 #    TestIt.script_module   string name
 ## TODO: template => new name
-import mezcla.template as THE_MODULE
+THE_MODULE = None                       ## TODO: remove
+## TODO: import mezcla.<module> as THE_MODULE
+#
+# Note: sanity test for customization (TODO: remove if desired)
+if not re.search(__file__, r"\btemplate.py$"):
+    debug.assertion("mezcla.template" not in str(THE_MODULE))
+
+## TODO:
+## # Environment options
+## # Note: These are just intended for internal options, not for end users.
+## # It also allows for enabling options in one place.
+## #
+## FUBAR = system.getenv_bool("FUBAR", False,
+##                            description="Fouled Up Beyond All Recognition processing")
 
 
 class TestIt(TestWrapper):
     """Class for testcase definition"""
-    script_module = TestWrapper.derive_tested_module_name(__file__)
+    script_module = TestWrapper.get_testing_module_name(__file__)
     # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
     # note: temp_file defined by parent (along with script_module, temp_base, and test_num)
 
@@ -43,15 +59,18 @@ class TestIt(TestWrapper):
     ## def setUpClass(cls):
     ##     """One-time initialization (i.e., for entire class)"""
     ##     debug.trace(6, f"TestIt.setUpClass(); cls={cls}")
-    ##     super(TestIt, cls).setUpClass()
+    ##     # note: should do parent processing first
+    ##     super().setUpClass()
     ##     ...
     ##     return
     ##
     ## def setUp(self):
     ##     """Per-test setup"""
-    ##     debug.trace(6, f"TestIt.setUp(); self={self}", 6)
-    ##     super(TestIt, self).setUp()
+    ##     debug.trace(6, f"TestIt.setUp(); self={self}")
+    ##     # note: must do parent processing first (e.g., for temp file support)
+    ##     super().setUp()
     ##     ...
+    ##     # TODO: debug.trace_current_context(level=debug.QUITE_DETAILED)
     ##     return
 
     def test_data_file(self):
