@@ -7,6 +7,14 @@
 # more classifiers, panda_sklearn.py should be used, espcially as it includes
 # Tensorflow and Keras.
 #
+# NOTES:
+# - Calculates accuracy as number of agreements over number of cases:
+#   ex:    (tp + tn) / (tp + tn + fp + fn)    for binary classifications
+# - See https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers#Single_metrics.
+# - Keep changes in sync with text_categorizer.py (e.g., XGBoost and GPU options).
+# - CherryPy Web server based on following tutorial
+#     https://simpletutorials.com/c/2165/How%20to%20Create%20a%20Simple%20JSON%20Service%20with%20CherryPy
+#
 # TODO:
 # - Maintain cache of categorization results.
 # - Review categorization code and add examples for clarification of parameters.
@@ -17,18 +25,16 @@
 """Text categorization support"""
 
 # Standard packages
-## OLD: import json
+from itertools import zip_longest
 import os
 import re
 import sys
-from collections import defaultdict
 
 # Installed packages
 import cherrypy
 import numpy
 import pandas
 from sklearn.base import BaseEstimator, ClassifierMixin
-## OLD: from sklearn.feature_extraction.text import CountVectorizer, TfidfTransformer
 from sklearn.feature_extraction.text import TfidfVectorizer
 from sklearn.feature_extraction.text import _document_frequency
 from sklearn.naive_bayes import MultinomialNB
@@ -38,19 +44,19 @@ from sklearn.svm import SVC
 from sklearn.pipeline import Pipeline
 from sklearn import metrics
 from sklearn.utils.multiclass import unique_labels
-## import xgboost as xgb
 
 # Local packages
 from mezcla import debug
-import mezcla.glue_helpers as gh
-import mezcla.misc_utils as misc
+from mezcla import glue_helpers as gh
+from mezcla import misc_utils as misc
 from mezcla import system
-from mezcla.system import getenv_bool, getenv_float, getenv_int, getenv_text
+from system import getenv_bool, getenv_float, getenv_int, getenv_text
 
 #................................................................................
 # Constants (e.g., environment-based options)
 
-SERVER_PORT = system.getenv_integer("SERVER_PORT", 9440)
+SERVER_PORT = system.getenv_integer("SERVER_PORT", 9010,
+                                    "TCP port for web interface")
 OUTPUT_BAD = system.getenv_bool("OUTPUT_BAD", False)
 CONTEXT_LEN = system.getenv_int("CONTEXT_LEN", 512)
 VERBOSE = system.getenv_bool("VERBOSE", False)
@@ -64,6 +70,8 @@ RETAIN_LABELS = getenv_bool("RETAIN_LABELS", False,
                             "Don't encode class labels")
 ENCODE_CLASSES = getenv_bool("ENCODE_CLASSES", not RETAIN_LABELS,
                              "Encode classes using enumeration")
+TRACE_IMPORTANCES = getenv_bool("TRACE_IMPORTANCES", False,
+                                "Trace feature importances")
 
 # Options for Support Vector Machines (SVM)
 #
@@ -88,21 +96,24 @@ SGD_PENALTY = system.getenv_text("SGD_PENALTY", "l2")
 SGD_ALPHA = system.getenv_float("SGD_ALPHA", 0.0001)
 SGD_SEED = system.getenv_float("SGD_SEED", None)
 SGD_MAX_ITER = system.getenv_int("SGD_MAX_ITER", 5)
-## OLD" SGD_TOLERANCE = system.getenv_float("SGD_TOLERANCE", None)
+## OLD: SGD_TOLERANCE = system.getenv_float("SGD_TOLERANCE", None)
 SGD_VERBOSE = system.getenv_bool("SGD_VERBOSE", False)
 
 # Options for Extreme Gradient Boost (XGBoost)
 USE_XGB = system.getenv_bool("USE_XGB", False)
 if USE_XGB:
+    # pylint: disable=import-outside-toplevel, import-error
     import xgboost as xgb
+XGB_BOOSTER = system.getenv_value("XGB_BOOSTER", None)
 XGB_USE_GPUS = system.getenv_bool("XGB_USE_GPUS", False)
+XGB_VERBOSITY = getenv_int("XGB_VERBOSITY", 0, "Degree of verbosity from 0 to 3")
 
 # Options for Logistic Regression (LR)
 # TODO: add regularization
 USE_LR = system.getenv_bool("USE_LR", False)
 
 # Options for GPU usage
-GPU_DEVICE = system.getenv_text("GPU_DEVICE", "",
+GPU_DEVICE = system.getenv_value("GPU_DEVICE", "",
                                 "Device number for GPU (e.g., shown under nvidia-smi)")
 
 # Options for TFIDF transformation
@@ -111,7 +122,7 @@ GPU_DEVICE = system.getenv_text("GPU_DEVICE", "",
 # TODO: add others from sklearn/feature_extraction/text.py
 # ex: min/max_df
 TFIDF_MAX_TERMS = getenv_int("MAX_TERMS", None,
-                                    "Maximum number of terms in TF/IDF matrix")
+                             "Maximum number of terms in TF/IDF matrix")
 TFIDF_MIN_NGRAM = getenv_int("MIN_NGRAM_SIZE", None)
 TFIDF_MAX_NGRAM = getenv_int("MAX_NGRAM_SIZE", None)
 TFIDF_MIN_DF = getenv_float("MIN_DF", None)
@@ -124,27 +135,18 @@ all_use_settings = [USE_SVM, USE_SGD, USE_XGB, USE_LR]
 USE_NB = (not any(all_use_settings))
 debug.assertion(sum([int(use) for use in all_use_settings]) <= 1)
 
-# Globals
-# Note: just used for OUTPUT_CSV support.
-tfidf_vectorizer = None
-
 #................................................................................
 # Utility functions
 
 def sklearn_report(actual, predicted, actual_labels, predicted_labels, stream=sys.stdout):
     """Print classification analysis report for ACTUAL vs. PREDICTED indices with original LABELS and using STREAM"""
     stream.write("Performance metrics:\n")
-    ## BAD: stream.write(metrics.classification_report(actual, predicted, target_names=labels))
     indices = unique_labels(actual, predicted)
     labels = unique_labels(actual_labels, predicted_labels)
     stream.write(metrics.classification_report(actual, predicted,
                                                labels=indices, target_names=labels))
     stream.write("Confusion matrix:\n")
     # TODO: make showing all cases optional
-    ## BAD: possible_indices = range(len(labels))
-    ## BAD
-    ## possible_indices = list(range(len(labels)))
-    ## confusion = metrics.confusion_matrix(actual, predicted, possible_indices)
     confusion = metrics.confusion_matrix(actual, predicted,
                                          labels=indices)
     # TODO: make sure not clipped
@@ -169,6 +171,7 @@ def read_categorization_data(filename):
     debug.trace_fmtd(4, "read_categorization_data({f})", f=filename)
     labels = []
     values = []
+    # TODO: rework via system.open_file
     with open(filename) as f:
         for (i, line) in enumerate(f):
             line = system.from_utf8(line)
@@ -177,10 +180,8 @@ def read_categorization_data(filename):
                 labels.append(items[0].lower())
                 values.append(items[1])
             else:
-                ## OLD: debug.trace_fmtd(4, "Warning: Ignoring item w/ unexpected format at line {num}",
                 debug.trace_fmtd(4, "Warning: Ignoring item w/ unexpected format at line {num}: items: len={l} first={f} second={s}",  l=len(items), f=gh.elide(items[0]), s=gh.elide(items[0]),
                                  num=(i + 1))
-    ## OLD: debug.trace_fmtd(7, "table={t}", t=table)
     debug.trace_values(7, zip(labels, values), "table")
     return (labels, values)
 
@@ -218,6 +219,7 @@ class ClassifierWrapper(BaseEstimator, ClassifierMixin):
         """Constructor: records CLASSIFIER"""
         debug.trace_fmt(6, "{cl}.__init__(clf={c})", c=classifier, cl=str(type(self)))
         self.classifier = classifier
+        self.tfidf_vectorizer = None
 
     def _get_param_names(self):
         """Get parameter names for the estimator"""
@@ -232,36 +234,18 @@ class ClassifierWrapper(BaseEstimator, ClassifierMixin):
     def fit(self, training_x=None, training_y=None):
         """Delegates fit() invocation to classifier, after outputing CSV if desired"""
         if OUTPUT_CSV:
-            ## NOTE: save in pickle format for debugging
-            if debug.verbose_debugging():
-                system.save_object(BASENAME + ".x.csv.pickle", training_x)
-                system.save_object(BASENAME + ".y.csv.pickle", training_y)
-            ##
-            df_x = pandas.DataFrame(training_x.toarray())
-            df_y = pandas.DataFrame(training_y)
-            
-            ## HACK: use global pipeline to get feature names
-            def normalize(feature):
-                """Normalize feature name"""
-                return feature.replace(" ", "_")
-            ##
-            features = [normalize(f) for f in tfidf_vectorizer.get_feature_names()]
-            df_x.to_csv(BASENAME + ".x.csv.list", header=features, index=False)
-            df_y.to_csv(BASENAME + ".y.csv.list", header=[CLASS_VAR], index=False)
-            # Trace out the IDF values; pylint: disable=protected-access
-            ## TODO: debug.trace_value(6, zip(features, tfidf_vectorizer._idf_diag), "ngram idf's")
-            if debug.debugging(6):
-                df = _document_frequency(training_x)
-                sorted_features = sorted(zip(features, tfidf_vectorizer.idf_, df),
-                                         key=lambda f_idf: f_idf[1], reverse=True)
-                debug.trace_values(6, sorted_features, "ngram idf's & df's")
-            ## DEBUG:
-            debug.trace_object(6, tfidf_vectorizer)
+            self.output_csv(training_x, training_y, BASENAME)
+        if TRACE_IMPORTANCES:
+            debug.trace_fmt(1, "Feature importances: {imp}",
+                            imp=self.extract_feature_importance())
+        ## DEBUG: debug.trace_object(6, self.tfidf_vectorizer)
             
         return self.classifier.fit(training_x, training_y)
 
     def predict(self, sample):
         """Return predicted class for each SAMPLE (returning vector)"""
+        if OUTPUT_CSV:
+            self.output_csv(sample, (["n/a"] * sample.shape[0]), BASENAME + ".test")
         return self.classifier.predict(sample)
 
     def score(self, X, y, sample_weight=None):
@@ -271,7 +255,60 @@ class ClassifierWrapper(BaseEstimator, ClassifierMixin):
     def predict_proba(self, sample):
         """Return probabilities of outcome classes predicted for each SAMPLE (returning matrix)"""
         return self.classifier.predict_proba(sample)
+
+    def output_csv(self, x, y, basename):
+        """Output features in X and Y to BASENAME.csv"""
+        # TODO: move into TextCategorizer; put pickle and IDF support in separate method
+        debug.trace(6, f"output_csv(_, _, {basename}); self={self}")
+        debug.reference_var(self)
+
+        ## NOTE: save in pickle format for debugging
+        if debug.verbose_debugging():
+            system.save_object(basename + ".x.csv.pickle", x)
+            system.save_object(basename + ".y.csv.pickle", y)
+        ##
+        df_x = pandas.DataFrame(x.toarray())
+        df_y = pandas.DataFrame(y)
+        
+        ## HACK: use global pipeline to get feature names
+        def normalize(feature_name):
+            """Normalize FEATURE_NAME"""
+            return feature_name.replace(" ", "_")
+        ##
+        features = [normalize(f) for f in self.tfidf_vectorizer.get_feature_names()]
+        df_x.to_csv(basename + ".x.csv.list", header=features, index=False)
+        df_y.to_csv(basename + ".y.csv.list", header=[CLASS_VAR], index=False)
+        gh.run("paste --delimiters=',' {b}.x.csv.list {b}.y.csv.list > {b}.csv.list",
+               b=basename)
+        # TODO: combine into single dataframe and use to_csv over that; use .csv instead of .csv.list (and make sure not to overwrite)
+        debug.assertion(system.file_exists(basename + ".csv.list"))
+
+        # Trace out the IDF values
+        ## TODO: debug.trace_value(6, zip(features, tfidf_vectorizer._idf_diag), "ngram idf's")
+        if debug.debugging(6):
+            # pylint: disable=protected-access
+            df = _document_frequency(x)
+            sorted_features = sorted(zip(features, self.tfidf_vectorizer.idf_, df),
+                                     key=lambda f_idf: f_idf[1], reverse=True)
+            debug.trace_values(6, sorted_features, "ngram idf's & df's")
+        return
     
+    def extract_feature_importance(self):
+        """Returns list of (name, weight) for important features"""
+        # TODO: move into TextCategorizer
+        debug.trace(6, f"extract_feature_importance(); self={self}")
+        result = []
+        try:
+            feature_names = (self.tfidf_vectorizer.get_feature_names() or [])
+            sorted_scores = sorted(zip_longest(feature_names, self.classifier.feature_importances_,  fillvalue="F?"),
+                                   key=lambda name_score: name_score[1],
+                                   reverse=True)
+            result = [(f, s) for (f, s) in sorted_scores if s > 0]
+        except:
+            system.print_exception_info("extract_feature_importance")
+        debug.trace(5, f"extract_feature_importance() => {result}")
+        return result
+
 #...............................................................................
 
 class TextCategorizer(object):
@@ -306,8 +343,6 @@ class TextCategorizer(object):
                                        alpha=SGD_ALPHA,
                                        random_state=SGD_SEED,
                                        ## TODO: max_iter=SGD_MAX_ITER,
-                                       ## OLD: n_iter=SGD_MAX_ITER,
-                                       ## OLD: tol=SGD_TOLERANCE
                                        verbose=SGD_VERBOSE)
             ## HACK: support old version and new (thanks sklearn!)
             ## TODO: make sure this won't break (e.g., due to visibility)
@@ -320,16 +355,12 @@ class TextCategorizer(object):
             # TODO: rework to just define classifier here and then pipeline at end.
             # in order to eliminate redundant pipeline-specification code.
             # TODO: n_jobs=-1
-            misc_xgb_params = {}
+            misc_xgb_params = {'booster': XGB_BOOSTER, 'verbosity': XGB_VERBOSITY}
             if XGB_USE_GPUS:
                 misc_xgb_params.update({'tree_method': 'gpu_hist'})
                 misc_xgb_params.update({'predictor': 'gpu_predictor'})
             if GPU_DEVICE:
                 misc_xgb_params.update({'gpu_id': GPU_DEVICE})
-            ## OLD:
-            ## if not XGB_USE_GPUS:
-            ##    misc_xgb_params['n_gpus'] = 0
-            ## debug.trace_fmt(6, 'misc_xgb_params={m}', m=misc_xgb_params)
             debug.trace_fmt(4, 'misc_xgb_params={m}', m=misc_xgb_params)
             classifier = xgb.XGBClassifier(**misc_xgb_params)
         elif USE_LR:
@@ -337,9 +368,10 @@ class TextCategorizer(object):
         else:
             debug.assertion(USE_NB)
             classifier = MultinomialNB()
-        if OUTPUT_CSV:
+        use_classifier_wrapper = (OUTPUT_CSV or TRACE_IMPORTANCES)
+        if use_classifier_wrapper:
             classifier = ClassifierWrapper(classifier)
-            debug.trace_fmt(4, "Using wrapper ({cl}) for CSV hooks", cl=type(classifier))
+            debug.trace_fmt(4, "Using wrapper ({cl}) for feature tracing hooks", cl=type(classifier))
 
         # Add classifier to text categorization pipeline]
         tfidf_parameters = {}
@@ -352,27 +384,21 @@ class TextCategorizer(object):
             debug.assertion(1 <= min_ngram <= max_ngram)
             tfidf_parameters['ngram_range'] = (min_ngram, max_ngram)
         if self.tfidf_min_df:
-            ## old: tfidf_parameters['min_df'] = tfidf_min_df
             tfidf_parameters['min_df'] = int_if_whole(self.tfidf_min_df)
         if self.tfidf_max_df:
-            ## old: tfidf_parameters['max_df'] = tfidf_max_df
             tfidf_parameters['max_df'] = int_if_whole(self.tfidf_max_df)
         if not self.tfidf_stopwords:
             tfidf_parameters['stop_words'] = 'english'
         if self.tfidf_char_ngrams:
             tfidf_parameters['analyzer'] = 'char'
         self.cat_pipeline = Pipeline(
-            [## OLD:
-             ## ('vect', CountVectorizer()),
-             ## ('tfidf', TfidfTransformer()),
-             ## NOTE: TfidfVectorizer same as CountVectorizer plus TfidfTransformer.
-             ('tfidf', TfidfVectorizer(**tfidf_parameters)),
+            [('tfidf', TfidfVectorizer(**tfidf_parameters)),
              ('clf', classifier)])
-        if OUTPUT_CSV:
+        if use_classifier_wrapper:
             pipeline_steps = list(self.cat_pipeline._iter())
-            global tfidf_vectorizer
-            tfidf_vectorizer = pipeline_steps[0][2]
             debug.assertion(pipeline_steps[0][1] == 'tfidf')
+            ## TODO: classifier.tfidf_vectorizer = self.cat_pipeline['tfidf']
+            classifier.tfidf_vectorizer = pipeline_steps[0][2]
         debug.trace_object(5, self, "TextCategorizer")
         return
 
@@ -382,25 +408,22 @@ class TextCategorizer(object):
         (labels, values) = read_categorization_data(filename)
         label_values = labels
         self.keys = sorted(numpy.unique(labels))
+        debug.trace_expr(5, self.keys)
         if ENCODE_CLASSES:
             label_indices = [self.keys.index(l) for l in labels]
             label_values = label_indices
-        ## OLD: self.classifier = self.cat_pipeline.fit(values, label_indices)
         self.classifier = self.cat_pipeline.fit(values, label_values)
-        ## OLD: debug.trace_object(7, self.classifier, "classifier")
         debug.trace_object(7, self, "TextCategorizer")
         return
 
     def test(self, filename, report=False, stream=sys.stdout):
         """Test classifier over tabular data from FILENAME with label and text, returning accuracy. Optionally, a detailed performance REPORT is output to STREAM."""
         debug.trace_fmtd(4, "tc.test({f})", f=filename)
-        ## OLD: (labels, values) = read_categorization_data(filename)
         (all_labels, all_values) = read_categorization_data(filename)
         debug.trace_values(6, all_labels, "all_labels")
         debug.trace_values(6, [gh.elide(v) for v in all_values], "all_values")
 
         # Prune cases with classes not in training data
-        ## BAD: actual_indices = [self.keys.index(l) for l in labels]
         # TODO: use hash of positions
         actual_indices = []
         values = []
@@ -415,14 +438,12 @@ class TextCategorizer(object):
                                  l=label, n=(i + 1))
 
         # Perform classification and determine accuracy
-        ## OLD: predicted_indices = self.classifier.predict(values)
         predicted_values = self.classifier.predict(values)
         if ENCODE_CLASSES:
             predicted_indices = predicted_values
         else:
             predicted_indices = [self.keys.index(label) for label in predicted_values]
         debug.assertion(len(actual_indices) == len(predicted_indices))
-        ## BAD: debug.trace_fmt(5, "actual: {act}\npredct: {pred}\n", act=actual_indices, pred=predicted_indices)
         debug.trace_values(6, actual_indices, "actual")
         debug.trace_values(6, predicted_indices, "predicted")
         ## TODO: predicted_labels = [self.keys[i] for i in predicted_indices]
@@ -431,31 +452,31 @@ class TextCategorizer(object):
 
         # Output classification report
         if report:
+            debug.assertion(VERBOSE)
             if VERBOSE:
+                stream.write("Missed classifications")
                 stream.write("\n")
                 stream.write("Actual\tPredict\n")
-                ## OLD: for i in range(len(actual_indices)):
                 ## TODO: complete conversion to using actual_index (here and below)
+                num_missed = 0
                 for (i, actual_index) in enumerate(actual_indices):
                     debug.assertion(actual_index == actual_indices[i])
-                    stream.write("{act}\t{pred}\n".
-                                 format(act=self.keys[actual_indices[i]],
-                                        pred=self.keys[predicted_indices[i]]))
+                    if (actual_indices[i] != predicted_indices[i]):
+                        stream.write("{act}\t{pred}\n".
+                                     format(act=self.keys[actual_indices[i]],
+                                            pred=self.keys[predicted_indices[i]]))
+                        num_missed += 1
+                if (num_missed == 0):
+                    stream.write("n/a")
                 stream.write("\n")
-            ## BAD: sklearn_report(actual_indices, predicted_indices, self.keys, stream)
-            ## OLD: keys = sorted(numpy.unique(labels))
-            ## BAD: keys = self.keys
-            ## BAD: sklearn_report(actual_indices, predicted_indices, keys, stream)
             actual_labels = [self.keys[i] for i in actual_indices]
             predicted_labels = [self.keys[i] for i in predicted_indices]
-            
             sklearn_report(actual_indices, predicted_indices, actual_labels, predicted_labels, stream)
 
         # Show cases not classified OK
         if OUTPUT_BAD:
             bad_instances = "Actual\tBad\tText\n"
             # TODO: for (i, actual_index) in enumerate(actual_indices)
-            ## OLD: for i in range(len(actual_indices)):
             for (i, actual_index) in enumerate(actual_indices):
                 debug.assertion(actual_index == actual_indices[i])
                 if (actual_indices[i] != predicted_indices[i]):
@@ -466,7 +487,6 @@ class TextCategorizer(object):
                         g=self.keys[actual_indices[i]],
                         b=self.keys[predicted_indices[i]],
                         t=context)
-            ## OLD: system.write_file(filename + ".bad", bad_instances)
             bad_filename = filename + ".bad"
             system.write_file(bad_filename, bad_instances)
             debug.trace_fmt(4, "Result ({f}):\n{r}", f=bad_filename, r=system.read_file(bad_filename))
@@ -474,7 +494,6 @@ class TextCategorizer(object):
 
     def categorize(self, text):
         """Return category for TEXT"""
-        # TODO: Add support for category distribution
         debug.trace(4, "tc.categorize(_)")
         debug.trace_fmtd(6, "\ttext={t}", t=text)
         index = self.classifier.predict([text])[0]
@@ -486,12 +505,10 @@ class TextCategorizer(object):
         """Return probability distribution for TEXT"""
         debug.trace(4, "tc.class_probabilities(_)")
         debug.trace_fmtd(6, "\ttext={t}", t=text)
-        ## BAD: class_names = self.classifier.classes_
         class_names = self.keys
         class_probs = self.classifier.predict_proba([text])[0]
         debug.trace_object(7, self.classifier)
         debug.trace_fmtd(6, "class_names: {cn}\nclass_probs: {cp}", cn=class_names, cp=class_probs)
-        ## BAD: dist = str(zip(class_names, class_probs))
         sorted_scores = misc.sort_weighted_hash(dict(zip(class_names, class_probs)))
         dist=" ".join([(k + ": " + system.round_as_str(s)) for (k, s) in sorted_scores])
         debug.trace_fmtd(5, "class_probabilities() => {r}", r=dist)
@@ -517,62 +534,19 @@ class TextCategorizer(object):
 # CherryPy Web server based on following tutorial
 #     https://simpletutorials.com/c/2165/How%20to%20Create%20a%20Simple%20JSON%20Service%20with%20CherryPy
 #
-# TODO: move to ~/visual-diff (e.g., text_categorizer_server.py)
-#
 
 # Constants
-TRUMP_TEXT = "Donald Trump is President."
+TRUMP_TEXT = "Donald Trump was President."
 DOG_TEXT = "My dog has fleas."
 
-## OLD:
-## CATEGORY_IMAGE_HASH = {
-##     # TODO: just use <category>.png to eliminate the hash
-##     # NOTES:
-##     # - drugs conflates with health
-##     # - government conflated with politics
-##     # - pets conflated with animal
-##     "animal": "/static/animals.png",
-##     "art": "/static/art.png",
-##     "biology": "/static/science.png",
-##     "business": "/static/business.jpg",
-##     "computers": "/static/computers.jpg",
-##     "drugs": "/static/health.jpg",
-##     "economics": "/static/economics.jpg",
-##     "education": "/static/education.png",
-##     "engineering": "/static/engineering.jpg",
-##     "food": "/static/food.jpg",
-##     "geography": "/static/geography.png",
-##     "geometry": "/static/geometry.jpg",
-##     "government": "/static/politics.png",
-##     "health": "/static/health.jpg",
-##     "history": "/static/history.jpg",
-##     "internet": "/static/internet.jpg",
-##     "law": "/static/law.jpg",
-##     "mathematics": "/static/mathematics.jpg",
-##     "military": "/static/military.png",
-##     "movie": "/static/movie.jpg",
-##     "music": "/static/music.jpg",
-##     "news": "/static/news.png",
-##     "pets": "/static/animals.png",
-##     "philosophy": "/static/philosophy.jpg",
-##     "politics": "/static/politics.png",
-##     "psychology": "/static/psychology.png",
-##     "religion": "/static/religion.jpg",
-##     "science": "/static/science.png",
-##     "software": "/static/software.jpg",
-##     "sports": "/static/sports.jpg",
-##     "technology": "/static/technology.jpg",
-##     "television": "/static/television.jpg",
-##     "tools": "/static/tools.jpg",
-##     "weather": "/static/weather.png",
-## }
-
 #--------------------------------------------------------------------------------
 # Utility function(s)
 
 def format_index_html(base_url=None):
-    """Formats a simple HTML page illustrating the categorize and get_category_image API calls,
-    Note: BASE_URL provides the server URL (e.g., http://www.scrappycito.com:9440)"""
+    """Formats a simple HTML page illustrating the categorize and class_probabilities API calls,
+    Note: BASE_URL provides the server URL (e.g., http://www.my-categorizer.com:9999)"""
+    debug.trace(5, f"format_index_html(base_url={base_url})")
+    # TODO: parameterize template generation (e.g., to facilitate usage in derived classes of web_controller
     if (base_url is None):
         base_url = "http://127.0.0.1"
     if (base_url.endswith("/")):
@@ -581,33 +555,31 @@ def format_index_html(base_url=None):
     # Create index page template with optional examples for debugging
     html_template = """
     <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
-    <html>
+    <html lang="en">
         <head>
+            <meta content="text/html; charset=UTF-8" http-equiv="content-type">
             <title>Text categorizer</title>
         </head>
         <body>
-            Try <a href="categorize">categorize</a> and <a href="get_category_image">get_category_image</a>.<br>
+            Try <a href="categorize">categorize</a> and <a href="class_probabilities">class_probabilities</a>.<br>
             note: You need to supply the <i><b>text</b></i> parameter.<br>
             <br>
-            For example,
+            Examples:
             <ul>
                 <li>Category for <a href="categorize?text={quoted_trump_text}">"{trump_text}"</a>:<br>
                     {indent}<code>{base_url}/categorize?text={quoted_trump_text}</code>
                 </li>
     
-                <li>Image for <a href="get_category_image?text={quoted_dog_text}">"{dog_text}"</a>:<br>
-                    {indent}<code>{base_url}/get_category_image?text={quoted_dog_text}</code>
+                <li>Probability distribution for <a href="class_probabilities?text={quoted_dog_text}">"{dog_text}"</a>:<br>
+                    {indent}<code>{base_url}/class_probabilities?text={quoted_dog_text}</code>
                 </li>
             </ul>
     """
-## TODO: drop get_category_image above
-##
-    html_template = html_template.replace("get_category_image", "class_probabilities").replace("Image", "Probabilities")
-    #
+
     if debug.detailed_debugging():
         html_template += """
-            <p>
-            Other examples:
+            <!-- <p> -->
+            Other examples (n.b., debug only):
             <ul>
                 <li><a href="shutdown">Shutdown</a> the server:<br>
                     {indent}<code>{base_url}/shutdown</code>
@@ -619,22 +591,25 @@ def format_index_html(base_url=None):
             </ul>
         """
     #
+    # TODO: define text area dimensions based on browser window size
     html_template += """
-	    <!-- Form for entering text for categorization -->
+            <!-- Form for entering text for categorization -->
             <hr>
-	    <form action="http://localhost:9440/categorize" method="get">
-	        <label for="textarea1">Categorize</label>
-	        <textarea id="textarea1" multiline="True" rows="10" cols="132" name="text"></textarea>
-	        <br>
-	        <input type="submit">
-	    </form>
-	    
+            <form action="{base_url}/categorize" method="get">
+                <label for="textarea1">Categorize</label>
+                <br>
+                <textarea id="textarea1" rows="5" cols="100" name="text"></textarea>
+                <br>
+                <input type="submit">
+            </form>
+            
         </body>
     </html>
     """
 
     # Resolve template into final HTML
-    index_html = html_template.format(base_url=base_url, indent="&nbsp;&nbsp;&nbsp;&nbsp;",
+    index_html = html_template.format(base_url=base_url,
+                                      indent="&nbsp;&nbsp;&nbsp;&nbsp;",
                                       trump_text=TRUMP_TEXT,
                                       quoted_trump_text=system.quote_url_text(TRUMP_TEXT),
                                       dog_text=DOG_TEXT,
@@ -646,7 +621,6 @@ def format_index_html(base_url=None):
 
 class web_controller(object):
     """Controller for CherryPy web server with embedded text categorizer"""
-    # TODO: put visual-diff support in ~/visual-diff directory (e.g., category image mapping)
     
     def __init__(self, model_filename, *args, **kwargs):
         """Class constructor: initializes search engine server"""
@@ -654,23 +628,13 @@ class web_controller(object):
                          s=self, a=args, k=kwargs)
         self.text_cat = TextCategorizer()
         self.text_cat.load(model_filename)
-        self.category_image = defaultdict(lambda: "/static/unknown-with-question-marks.png")
-        ## OLD:
-        ## # HACK: wikipedia categorization specific
-        ## self.category_image.update(CATEGORY_IMAGE_HASH)
-        # Note: To avoid cross-origin type errrors, Access-Control-Allow-Origin
-        # is made open. See following:
-        # - http://cleanbugs.com/item/how-to-get-cross-origin-sharing-cors-post-request-working-a-resource-413656.html
-        # - https://stackoverflow.com/questions/6054473/python-cherrypy-how-to-add-header
-        # TODO: put cherrypy config in start_web_controller (or put it's configuration here)
-        ## BAD: cherrypy.response.headers["Access-Control-Allow-Origin"] = "*"
         return
 
     @cherrypy.expose
     def index(self, **kwargs):
         """Website root page (e.g., web site overview and link to search)"""
+        # TODO: add way to override URL (e.g., to force use of known local hostname instead of "localhost")
         debug.trace_fmtd(5, "wc.index(s:{s}, kw:{kw})", s=self, kw=kwargs)
-        ## OLD: return "not much here excepting categorize and get_category_image"
         base_url = cherrypy.url('/')
         debug.trace_fmt(4, "base_url={b}", b=base_url)
         index_html = format_index_html(base_url)
@@ -691,45 +655,17 @@ class web_controller(object):
     #
     probs = class_probabilities
 
-    ## OLD
-    ## @cherrypy.expose
-    ## ## @cherrypy.tools.json_out()
-    ## def get_category_image(self, text, **kwargs):
-    ##     """Infer category for TEXT and return image"""
-    ##     debug.trace_fmtd(5, "wc.get_category_image(_, {kw}); self={s}", t=text, s=self, kw=kwargs)
-    ##     ## TEST: debug.trace_fmtd(6, "\ttext={t}", t=text)
-    ##     cat = self.categorize(text, **kwargs)
-    ##     image = self.category_image[cat]
-    ##     # for JSONP, need to add callback call and format the call
-    ##     # TODO: see if cherrypy handles this
-    ##     # see https://stackoverflow.com/questions/19456146/ajax-call-and-clean-json-but-syntax-error-missing-before-statement
-    ##     ## return image
-    ##     ## return json.dumps({"image": image})
-    ##     ## return {"image": image}
-    ##     image_id = kwargs.get("id", "id0")
-    ##     result = json.dumps({"image": image, "id": image_id})
-    ##     if 'callback' in kwargs:
-    ##         callback_function = kwargs['callback']
-    ##         debug.trace_fmtd(5, "Invoking callback {cb}", cb=callback_function)
-    ##         data = kwargs.get("data", "")
-    ##         result = (callback_function + "(" + result + ", " + data + ");")
-    ##     ## OLD: debug.trace_fmtd(6, "wc.get_category_image() => {r}", r=result)
-    ##     debug.trace_fmtd(6, "wc.get_category_image({t}) => {r}; cat={c}", t=text, r=result, c=cat)
-    ##     return result
-
     @cherrypy.expose
     def stop(self, **kwargs):
         """Stops the web search server and saves cached data to disk.
         Note: The command is ignored if not debugging."""
         debug.trace_fmtd(5, "wc.stop(s:{s}, kw:{kw})", s=self, kw=kwargs)
-        # TODO: get whitelisted server hosts from environment
-        ## OLD: if ((not debug.detailed_debugging()) and (os.environ.get("HOST_NICKNAME") in ["hostwinds", "hw2", "ec2-micro"])):
-        # TODO: replace stooges with real server nicknames
+        # TODO: replace stooges with your real server nicknames
         if ((not debug.detailed_debugging()) and (os.environ.get("HOST_NICKNAME") in ["curly", "larry", "moe"])):
             return "Call security!"
         # TODO: Straighten out shutdown quirk (seems like two invocations required).
         # NOTE: Putting exit before stop seems to do the trick. However, it might be
-        # the case that the servr shutdown 
+        # the case that the server shutdown.
         cherrypy.engine.exit()
         cherrypy.engine.stop()
         # TODO: Use HTML so shutdown shown in title.
@@ -741,8 +677,8 @@ class web_controller(object):
 
 
 def start_web_controller(model_filename):
-    """Start up the CherryPy controller for categorization via MODEL_FILENAME"""
-    # TODO: return status code
+    """Start up the CherryPy controller for categorization via MODEL_FILENAME
+    Note: The function blocks until server is shutdown"""
     debug.trace(5, "start_web_controller()")
 
     # Load in CherryPy configuration
@@ -751,10 +687,10 @@ def start_web_controller(model_filename):
         '/': {
             'tools.sessions.on': True,
             'tools.staticdir.root': os.path.abspath(os.getcwd()),
-            ## take 2: on avoiding cross-origin type errrors
+            ## notes: avoids cross-origin type errrors
             'tools.response_headers.on': True,
             'tools.response_headers.headers': [
-                ## OLD: ('Content-Type', 'text/javascript'),
+                ## TODO: just allow the same host
                 ('Access-Control-Allow-Origin', '*'),
             ]
         },
@@ -768,8 +704,10 @@ def start_web_controller(model_filename):
     # Start the server
     # TODO: trace out all configuration settings
     debug.trace_values(4, cherrypy.response.headers, "default response headers")
-    cherrypy.quickstart(web_controller(model_filename), "", conf)
-    ## TODO: debug.trace_value(4, cherrypy.response.headers, "response headers")
+    textcat_controller = web_controller(model_filename)
+    debug.trace_expr(4, textcat_controller.text_cat.keys)
+    cherrypy.quickstart(textcat_controller, config=conf)
+    # Note: the following call blocks
     cherrypy.engine.start()
     return
 
@@ -780,7 +718,10 @@ def start_web_controller(model_filename):
 def main(args):
     """Supporting code for command-line processing"""
     debug.trace_fmtd(6, "main({a})", a=args)
-    if (len(args) != 2):
+    # HACK: ignore --tag label (n.b., used for killing via process regex)
+    if ((len(args) > 0) and (args[1] == "--tag")):
+        args[1:] = args[3:]
+    if ((len(args) != 2) or (args[1] == "--help")):
         system.print_stderr("Usage: {p} model".format(p=args[0]))
         return
     model = args[1]
