@@ -9,12 +9,24 @@
 # - This is in the spirit of quick-n-dirty (e.g., for R&D: they are packages
 #   that are better suited for industrial strength code (e.g., for production).
 #
+# Usage examples:
+#
+#   from mezcla import debug
+#   from mezcla import text_utils
+#
+#   import tensorflow
+#   from mezcla.text_utils import version_to_number as version2num
+#   debug.assertion(version2num(tensorflow.__version__) <= version2num("1.15"))
+#
+#   devlin_text = document_to_text("devlin-bert-pretraining-arxiv-1810-04805-dec19.pdf")
+#   str(devlin_text).count("BERT")
+#   => 117
+#
 # TODO:
 # - Write up test suite, el tonto!.
 # - Add pointer to specific packages better for production use.
 # - Move HTML-specific functions into html_utils.py.
 #
-#
 
 """Miscellaneous text utility functions"""
 
@@ -30,6 +42,7 @@ import sys
 
 # Local packages
 from mezcla import debug
+from mezcla import glue_helpers as gh
 from mezcla.my_regex import my_re
 from mezcla import system
 from mezcla.system import to_int
@@ -50,6 +63,7 @@ def html_to_text(document_data):
     # EX: html_to_text("<html><body><!-- a cautionary tale -->\nMy <b>fat</b> dog has fleas</body></html>") => "My fat dog has fleas"
     # Note: stripping javascript and style sections based on following:
     #   https://stackoverflow.com/questions/22799990/beatifulsoup4-get-text-still-has-javascript
+    # TODO: move into html_utils.py
     debug.trace_fmtd(7, "html_to_text(_):\n\tdata={d}", d=document_data)
     ## OLD: soup = BeautifulSoup(document_data)
     init_BeautifulSoup()
@@ -60,8 +74,9 @@ def html_to_text(document_data):
         # -or- Note the in-place change (i.e., destructive).
         script.extract()
     # Get the text
-    text = soup.get_text()
-    debug.trace_fmtd(6, "html_to_text() => {t}", t=text)
+    ## OLD: text = soup.get_text()
+    text = soup.get_text(separator=" ")
+    debug.trace_fmtd(6, "html_to_text() => {t}", t=gh.elide(text))
     return text
 
 
@@ -77,10 +92,15 @@ def document_to_text(doc_filename):
     text = ""
     try:
         init_textract()
-        text = system.from_utf8(textract.process(doc_filename))
+        ## OLD: text = system.from_utf8(textract.process(doc_filename))
+        text = textract.process(doc_filename)
+        if isinstance(text, bytes):
+            text = text.decode("UTF-8", errors="ignore")
     except:
         debug.trace_fmtd(3, "Warning: problem converting document file {f}: {e}",
                          f=doc_filename, e=sys.exc_info())
+    debug.trace_fmt(5, "document_to_text({fn}) => {r}",
+                    fn=doc_filename, r=gh.elide(text))
     return text
 
 
@@ -89,6 +109,7 @@ def extract_html_images(document_data, url):
     debug.trace_fmtd(8, "extract_html_images(_):\n\tdata={d}", d=document_data)
     # TODO: add example; return dimensions
     # TODO: have URL default to current directory
+    # TODO: put in html_utils
 
     # Parse HTML, extract base URL if given and get website from URL.
     init_BeautifulSoup()
@@ -115,6 +136,7 @@ def extract_html_images(document_data, url):
 
     # Get images and resolve to full URL (TODO: see if utility for this)
     # TODO: include CSS background images
+    # TODO: use DATA-SRC if SRC not valid URL (e.g., src="data:image/gif;base64,R0lGODl...")
     images = []
     all_images = soup.find_all('img')
     for image in all_images:
@@ -131,7 +153,10 @@ def extract_html_images(document_data, url):
             image_src = web_site_url + image_src
         elif not image_src.startswith("http"):
             image_src = base_url + "/" + image_src
-        images.append(image_src)
+        ## TEMP: fixup for trailing newline (TODO: handle upstream)
+        image_src = image_src.strip()
+        if image_src not in images:
+            images.append(image_src)
     debug.trace_fmtd(6, "extract_html_images() => {i}", i=images)
     return images
 
@@ -142,6 +167,9 @@ def version_to_number(version, max_padding=3):
     major number m is the minor, r is the revision and o is other.
     Each version component will be prepended with up MAX_PADDING [3] 0's
     Notes:
+    - comparisons should be made against using functional constant, not number:
+      EX: version_to_number("1.1.1") > version_to_number("1.1")
+      EX: version_to_number("1.1.1") < 1.1
     - strings in the version are ignored
     - 0 is returned if version string is non-standard"""
     # EX: version_to_number("1.11.1") => 1.00010001
@@ -307,5 +335,5 @@ def make_fixed_length(text, length=16):
 #-------------------------------------------------------------------------------
 
 if __name__ == '__main__':
-    system.print_stderr("Error: not intended for command-line use")
+    system.print_stderr("Warning: not intended for command-line use")
     debug.assertion("html" not in html_to_text("<html><body></body></html>"))
