@@ -5,7 +5,7 @@
 # Notes:
 # - Based on template.py used in older test scripts
 # - Creates per-test temp file, based on same class-wide temp-base file.
-# - To treat temp-base as a subdirectory. set use_temp_base_dir to True in 
+# - To treat temp-base as a subdirectory, set use_temp_base_dir to True in 
 #   class member initialiation section.
 # - Changes to temporary directory/file should be synchronized with ../main.py.
 # - Overriding the temporary directory can be handy during debugging; however,
@@ -15,10 +15,9 @@
 #-------------------------------------------------------------------------------
 # Sample test (streamlined version of test_simple_main_example.py):
 #
-#   import unittest 
-#   from unittest_wrapper import TestWrapper
-#   import mezcla.glue_helpers as gh
-#   ## TODO: import sample_test as THE_MODULE
+#   import unittest
+#   from mezcla import system
+#   from mezcla.unittest_wrapper import TestWrapper
 #
 #   class TestIt(TestWrapper):
 #       """Class for testcase definition"""
@@ -26,8 +25,8 @@
 #  
 #       def test_simple_data(self):
 #           """Make sure simple data sample processed OK"""
-#           gh.write_lines(self.temp, "really fubar")
-#           output = self.run_script("--fubar")
+#           system.write_file(self.temp_file, "really fubar")
+#           output = self.run_script("--check-fubar", self.temp_file)
 #           self.assertTrue("really" in output)
 #
 #   if __name__ == '__main__':
@@ -35,6 +34,7 @@
 #-------------------------------------------------------------------------------
 # TODO:
 # - Add method to invoke unittest.main(), so clients don't need to import unittest.
+# - Clarify how ALLOW_SUBCOMMAND_TRACING affects tests that invoke external scripts.
 #
 
 """Unit test support class"""
@@ -45,9 +45,12 @@ import tempfile
 import unittest
 
 import mezcla
-import mezcla.glue_helpers as gh
-import mezcla.tpo_common as tpo
-import mezcla.debug as debug
+from mezcla import debug
+from mezcla import glue_helpers as gh
+from mezcla import system
+from mezcla import tpo_common as tpo
+
+TODO_MODULE = "TODO MODULE"
 
 # Note: the following is for transparent resolution of dotted module names
 # for invocation of scripts via 'python -m package.module'. This is in support
@@ -60,9 +63,12 @@ debug.assertion(THIS_PACKAGE == "mezcla")
 
 class TestWrapper(unittest.TestCase):
     """Class for testcase definition"""
-    script_module = "TODO MODULE"        # name for invocation via 'python -m'
-    temp_base = tpo.getenv_text("TEMP_BASE",
-                                tempfile.NamedTemporaryFile().name)
+    script_module = TODO_MODULE         # name for invocation via 'python -m' (n.b., usuually set via derive_tested_module_name)
+    temp_base = system.getenv_text("TEMP_BASE",
+                                   tempfile.NamedTemporaryFile().name)
+    ## TODO: temp_file = None
+    ## TEMP: initialize to unique value independent of temp_base
+    temp_file = tempfile.NamedTemporaryFile().name
     use_temp_base_dir = None
     test_num = 1
 
@@ -71,17 +77,25 @@ class TestWrapper(unittest.TestCase):
         """Per-class initialization: make sure script_module set properly"""
         tpo.debug_format("TestWrapper.setupClass(): cls={c}", 5, c=cls)
         super(TestWrapper, cls).setUpClass()
-        ## OLD: tpo.trace_object(cls, 5, "TestWrapper class")
         debug.trace_object(5, cls, "TestWrapper class")
-        debug.assertion("TODO " not in cls.script_module)
+        debug.assertion(cls.script_module != TODO_MODULE)
         if cls.script_module:
+            # Try to pull up usage via python -m mezcla.xyz --help
             help_usage = gh.run("python -m '{mod}' --help", mod=cls.script_module)
-            gh.assertion("No module named" not in help_usage)
+            debug.assertion("No module named" not in help_usage,
+                            f"problem running via 'python -m {cls.script_module}'")
+            # Warn about lack of usage statement unless "not intended for command-line" type warning issued
+            # OLD: (re.search(r"not intended.*(command|standalone)", help_usage))
+            # TODO: standardize the not-intended wording
+            if (not ((re.search(r"warning:.*not intended", help_usage,
+                                re.IGNORECASE))
+                     or ("usage:" in help_usage.lower()))):
+                system.print_stderr("Warning: mezcla scripts should implement --help")
 
         # Optionally, setup temp-base directory (normally just a file)
         if cls.use_temp_base_dir is None:
-            cls.use_temp_base_dir = tpo.getenv_bool("USE_TEMP_BASE_DIR", False)
-            # TODO: temp_base_dir = tpo.getenv_text("TEMP_BASE_DIR", ""); cls.use_temp_base_dir = bool(temp_base_dir.strip); ...
+            cls.use_temp_base_dir = system.getenv_bool("USE_TEMP_BASE_DIR", False)
+            # TODO: temp_base_dir = system.getenv_text("TEMP_BASE_DIR", " "); cls.use_temp_base_dir = bool(temp_base_dir.strip()); ...
         if cls.use_temp_base_dir:
             ## TODO: pure python
             gh.run("mkdir -p {dir}", dir=cls.temp_base)
@@ -90,18 +104,23 @@ class TestWrapper(unittest.TestCase):
 
     @staticmethod
     def derive_tested_module_name(test_filename):
-        """Derive the name of the module being tested from TEST_FILENAME"""
-        # Note: Deprecated method
+        """Derive the name of the module being tested from TEST_FILENAME. Used as follows:
+              script_module = TestWrapper.derive_tested_module_name(__file__)
+        Note: Deprecated method (see get_testing_module_name)
+        """
+        tpo.debug_format("Warning: in deprecrated derive_tested_module_name", 5)
         module = os.path.split(test_filename)[-1]
         module = re.sub(r".py[oc]?$", "", module)
         module = re.sub(r"^test_", "", module)
-        tpo.debug_format("derive_tested_module_name({f}) = {m}", 5,
+        tpo.debug_format("derive_tested_module_name({f}) => {m}", 5,
                          f=test_filename, m=module)
         return (module)
 
     @staticmethod
-    def get_testing_module_name(test_filename, module_object):
+    def get_testing_module_name(test_filename, module_object=None):
         """Derive the name of the module being tested from TEST_FILENAME and MODULE_OBJECT"""
+        # TODO: used as follows (see tests/test_template.py):
+        #    script_module = TestWrapper.get_testing_module_name(__file__)
         module_name = os.path.split(test_filename)[-1]
         module_name = re.sub(r".py[oc]?$", "", module_name)
         module_name = re.sub(r"^test_", "", module_name)
@@ -109,16 +128,19 @@ class TestWrapper(unittest.TestCase):
         if module_object is not None:
            package_name = getattr(module_object, "__package__", "")
         full_module_name = package_name + "." + module_name
-        tpo.debug_format("derive_tested_module_name({f}) = {m}", 5,
+        tpo.debug_format("get_testing_module_name({f}) => {m}", 5,
                          f=test_filename, m=full_module_name)
         return (full_module_name)
 
     def setUp(self):
-        """Per-test initializations: disables tracing scripts invoked via run();
-        initializes temp file name (With override from environment)."""
+        """Per-test initializations
+        Notes:
+        - Disables tracing scripts invoked via run() unless ALLOW_SUBCOMMAND_TRACING
+        - Initializes temp file name (With override from environment)."""
         # Note: By default, each test gets its own temp file.
         tpo.debug_print("TestWrapper.setUp()", 4)
-        gh.disable_subcommand_tracing()
+        if not gh.ALLOW_SUBCOMMAND_TRACING:
+            gh.disable_subcommand_tracing()
         # The temp file is an extension of temp-base file by default.
         # Opitonally, if can be a file in temp-base subdrectory.
         if self.use_temp_base_dir:
@@ -126,35 +148,41 @@ class TestWrapper(unittest.TestCase):
         else:
             default_temp_file = self.temp_base + "-test-"
         default_temp_file += str(TestWrapper.test_num)
-        self.temp_file = tpo.getenv_text("TEMP_FILE", default_temp_file)
+        self.temp_file = system.getenv_text("TEMP_FILE", default_temp_file)
+        gh.delete_existing_file(self.temp_file)
         TestWrapper.test_num += 1
 
         ## OLD: tpo.trace_object(self, 5, "TestWrapper instance")
         debug.trace_object(5, self, "TestWrapper instance")
         return
 
-    def run_script(self, options=None, data_file=None, log_file=None, 
-                   out_file=None, env_options=""):
+    def run_script(self, options=None, data_file=None, log_file=None, trace_level=4,
+                   out_file=None, env_options=None, uses_stdin=False):
         """Runs the script over the DATA_FILE (optional), passing (positional)
         OPTIONS and optional setting ENV_OPTIONS. If OUT_FILE and LOG_FILE are
         not specifed, they  are derived from self.temp_file.
-        Note: issues warning if script invocation leads to error"""
-        tpo.debug_format("TestWrapper.run_script({opts}, {df}, {lf}, {of}, {env})", 5,
+        Notes:
+        - issues warning if script invocation leads to error
+        - if USES_STDIN, requires explicit empty string for DATA_FILE to avoid use of - (n.b., as a precaution against hangups)"""
+        tpo.debug_format("TestWrapper.run_script({opts}, {df}, {lf}, {of}, {env})", trace_level + 1,
                          opts=options, df=data_file, lf=log_file, 
                          of=out_file, env=env_options)
         if options is None:
             options = ""
+        if env_options is None:
+            env_options = ""
 
         # Derive the full paths for data file and log, and then invoke script.
-        # TODO: derive from temp base and data file name?
-        data_path = ""
-        if data_file:
-            data_path = gh.resolve_path(data_file)
+        # TODO: derive from temp base and data file name?;
+        data_path = ("" if uses_stdin else "-")
+        if data_file is not None:
+            data_path = (gh.resolve_path(data_file) if len(data_file) else data_file)
         if not log_file:
             log_file = self.temp_file + ".log"
         if not out_file:
             out_file = self.temp_file + ".out"
         # note: output is redirected to a file to preserve tabs
+        debug.assertion(not self.script_module.endswith(".py"))
         gh.issue("{env} python  -m {module}  {opts}  {path} 1> {out} 2> {log}",
                  env=env_options, module=self.script_module, 
                  opts=options, path=data_path, out=out_file, log=log_file)
@@ -162,7 +190,7 @@ class TestWrapper(unittest.TestCase):
         # note; trailing newline removed as with shell output
         if output.endswith("\n"):
             output = output[:-1]
-        tpo.debug_format("output: {{\n{out}\n}}", 4,
+        tpo.debug_format("output: {{\n{out}\n}}", trace_level,
                          out=gh.indent_lines(output))
 
         # Make sure no python or bash errors. For example,
@@ -171,7 +199,7 @@ class TestWrapper(unittest.TestCase):
         error_found = re.search(r"(\S+error:)|(no module)|(command not found)",
                                 log_contents.lower())
         gh.assertion(not error_found)
-        tpo.debug_format("log contents: {{\n{log}\n}}", 5,
+        tpo.debug_format("log contents: {{\n{log}\n}}", trace_level + 1,
                          log=gh.indent_lines(log_contents))
 
         # Do sanity check for python exceptions
